//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TContextView
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop

//#pragma option -Jgx

#include <owl\menu.h>

#include "views.rc"
#include "tree.h"
#include "strucdoc.h"
#include "mydc.h"
#include "viewnode.h"
#include "viewtree.h"
#include "scrwview.h"
#include "scrtrwvw.h"



class   TContextViewNode : public TViewNode {
public:
								 TContextViewNode(int n = 0, int fixed = FALSE);
	inline TContextViewNode *GetChild(int n) const;
	void						 CalcSizeString(TMyDC&, const char const *);
	void						 DrawString(TMyDC&, TPoint&, const char const *);
	void						 CalcSizeHole(TMyDC&);
	void						 DrawHole(TMyDC&, TPoint&);
	void  					 CalcSizeDelimiters(TMyDC& dc, \
														 char *start, char *between, char *end);

	void  					 DrawDelimiters(TMyDC& dc, TPoint& pos, \
														 char *start, char *between, char *end);
	static TNodeAddress   ActiveAddress;
};

inline TContextViewNode* TContextViewNode::GetChild(int n) const
{  return (TContextViewNode*)TTreeNode::GetChild(n);
}


const int         Indent  = 20;
const int         DistX   = 4;
const int         DistY   = 4;
#define  FocusColor   TColor(128, 128, 128)
static TBrush             FocusBrush  = TBrush(FocusColor);

TNodeAddress	 				 TContextViewNode::ActiveAddress;

								 TContextViewNode::TContextViewNode(int n, int fixed):
	TViewNode(n, fixed)
{
}
void				    TContextViewNode::CalcSizeString(TMyDC& dc, const char const *s)
{
	SetNodeSize(dc.GetTextExtent(s, strlen(s)));
}

void				    		 TContextViewNode::DrawString(TMyDC& dc, TPoint& pos, const char const *s)
{
	dc.TextOut(pos, s);
}


void						TContextViewNode::CalcSizeHole(TMyDC& dc)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	CalcSizeString(dc, s);
}

void						TContextViewNode::DrawHole(TMyDC& dc, TPoint& pos)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	DrawString(dc, pos, s);
}

void  TPROGHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "No context"); }
void  TPROGHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "No context"); }
void  TDECHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TDECSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TNDECHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TNDECHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TIDHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TIDHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TTYPEHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TTYPEHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TSTATHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TSTATSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TVARSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TVARSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS1HOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS1HOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS2HOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS2HOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALT2HOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALT2HOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPBODYHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPBODYHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFBODYHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFBODYHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFARGSHOLE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFARGSHOLE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }


void  TPROG<TContextViewNode>::CalcSize(TMyDC& dc)
{
	TSortDECS<TContextViewNode> 	*decs  = GetDeclarations(); decs->CalcSize(dc);
	TSize									 sz    = decs->NodeSize();
	TContextViewNode					*node  = this;
	int									 ind   = Indent;

	for(int i = 0; i<ActiveAddress.Length(); i++) {
		node = node->GetChild(ActiveAddress.GetChildNumber(i));
		Sort		sort = node->GetDesiredChildSort(0);
		if(sort==SORT_DECS || sort==SORT_PPARS || sort==SORT_FPARS) {
			TContextViewNode				*child = node->GetChild(0);
			child->CalcSize(dc);
			if(child->NodeHeight()) {
				sz.cy += child->NodeHeight()+DistY;
				sz.cx  = max(sz.cx,ind+child->NodeWidth());
				ind   += Indent;
			}
		}
	}
	SetNodeSize(sz);
}


void  TPROG<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	TSortDECS<TContextViewNode> 	*decs  = GetDeclarations();
	TPoint								 p = pos;
	TContextViewNode					*node  = this;

	decs->Draw(dc, p); p += TSize(Indent, DistY+decs->NodeHeight());
	for(int i = 0; i<ActiveAddress.Length(); i++) {
		node = node->GetChild(ActiveAddress.GetChildNumber(i));
		Sort		sort = node->GetDesiredChildSort(0);
		if(sort==SORT_DECS || sort==SORT_PPARS || sort==SORT_FPARS) {
			TContextViewNode				*child = node->GetChild(0);
			if(child->NodeHeight()) {
				child->Draw(dc, p);
				p += TSize(Indent, DistY+child->NodeHeight());
			}
		}
	}
}


void  TDECS<TContextViewNode>::CalcSize(TMyDC& dc)
{
		TSize    semicolonsize;
		dc.GetTextExtent(";",     1, semicolonsize);

		int width       = 0;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortDEC<TContextViewNode> *dec   = GetDeclaration(i);
			dec->CalcSize(dc);
			int sw = dec->NodeSize().cx+semicolonsize.cx;
			if(sw>width) width = sw;
			height += dec->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));

}
void  TDECS<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	int y = pos.y;
	int numchildren = GetNumChildren();
	TSize    semicolonsize;
	dc.GetTextExtent(";",     1, semicolonsize);

	for(int i = 0; i<numchildren; i++) {
		TSortDEC<TContextViewNode> *dec   = GetDeclaration(i);
		if(i) dc.TextOut(pos.x, y, ";");
		dec->Draw(dc, TPoint(pos.x+semicolonsize.cx, y));
		y += dec->NodeSize().cy;
	}
}


void  TContextViewNode::CalcSizeDelimiters(TMyDC& dc,
														 char *start, char *between, char *end)
{
		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int width       = startsize.cx;
		int height      = max(max(startsize.cy, betweensize.cy), endsize.cy);
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TContextViewNode *child   = GetChild(i);
			child->CalcSize(dc);
			width += child->NodeWidth() +(i<numchildren-1?betweensize.cx:0);
			if(child->NodeHeight() > height)
				height = child->NodeHeight();
		}
		width += endsize.cx;
		SetNodeSize(TSize(width, height));
}
void  TContextViewNode::DrawDelimiters(TMyDC& dc, TPoint& pos,
														 char *start, char *between, char *end)
{

		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int numchildren = GetNumChildren();

		if(start)dc.TextOut(pos, start);
		int x = pos.x + startsize.cx;
		for(int i = 0; i<numchildren; i++) {
			TContextViewNode *child   = GetChild(i);
			child->Draw(dc, TPoint(x, pos.y));
			x += child->NodeWidth();
			if(i<numchildren-1 && between) {
				dc.TextOut(x, pos.y, between); x += betweensize.cx;
			}
		}
		if(end)dc.TextOut(x, pos.y, end);

}


void  TNDECS<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "var  ", ", ", 0);
}
void  TNDECS<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "var  ", ", ", 0);
}
void  TNDEC<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc,  0, ":", 0);
}
void  TNDEC<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ":", 0);
}

void  TPROCDEC<TContextViewNode>::CalcSize(TMyDC& dc)
{
	TSortID<TContextViewNode>	 *id = GetProcName();
	id->CalcSize(dc);

	TSize    procsize, eqsize;
	dc.GetTextExtent("proc ", 5, procsize);

	SetNodeSize(TSize(procsize.cx + id->NodeWidth(), id->NodeHeight()));
}
void  TPROCDEC<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	TSize    procsize, eqsize;
	dc.GetTextExtent("proc ", 5, procsize);
	dc.TextOut(pos, "proc");
	int	x = pos.x+procsize.cx;
	TSortID<TContextViewNode>	 *id = GetProcName();
	id->Draw(dc, TPoint(x, pos.y));
}


void  TFUNCDEC<TContextViewNode>::CalcSize(TMyDC& dc)
{
	TSortID<TContextViewNode>	 *id = GetFuncName();
	id->CalcSize(dc);

	TSize    FUNCsize;
	dc.GetTextExtent("func ", 5, FUNCsize);

	SetNodeSize(TSize(FUNCsize.cx + id->NodeWidth(),
							id->NodeHeight()));
}
void  TFUNCDEC<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	TSize    FUNCsize;
	dc.GetTextExtent("func ", 5, FUNCsize);
	dc.TextOut(pos, "func");
	int	x = pos.x+FUNCsize.cx;
	TSortID<TContextViewNode>	 *id = GetFuncName();
	id->Draw(dc, TPoint(x, pos.y));
}


void  TID<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TID<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TINTTYPE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "int"); }
void  TINTTYPE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "int"); }
void  TCHARTYPE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "char"); }
void  TCHARTYPE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "char"); }
void  TREALTYPE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "real"); }
void  TREALTYPE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "real"); }
void  TBOOLTYPE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "bool"); }
void  TBOOLTYPE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "bool"); }

void  TSTATS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TSTATS<TContextViewNode>::Draw(TMyDC&, TPoint&){}

void  TSKIP<TContextViewNode>::CalcSize(TMyDC&) {}
void  TSKIP<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TCONCASSIGN<TContextViewNode>::CalcSize(TMyDC&) {}
void  TCONCASSIGN<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPROCCALL<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPROCCALL<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPROCCALLARGS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPROCCALLARGS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TIFTHEN<TContextViewNode>::CalcSize(TMyDC&) {}
void  TIFTHEN<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TIFTHENELSE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TIFTHENELSE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TWHILE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TWHILE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TREAD<TContextViewNode>::CalcSize(TMyDC&) {}
void  TREAD<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TWRITE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TWRITE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TFOR<TContextViewNode>::CalcSize(TMyDC&) {}
void  TFOR<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TBLOCK<TContextViewNode>::CalcSize(TMyDC&) {}
void  TBLOCK<TContextViewNode>::Draw(TMyDC&, TPoint&) {}


void  TVARS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TVARS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TEXPRS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TEXPRS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TVARVALUE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TVARVALUE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TFUNCCALL<TContextViewNode>::CalcSize(TMyDC&) {}
void  TFUNCCALL<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TFUNCCALLARGS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TFUNCCALLARGS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TNUMVALUE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TNUMVALUE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TCHARVALUE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TCHARVALUE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TIMPLICATION<TContextViewNode>::CalcSize(TMyDC&) {}
void  TIMPLICATION<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TEQUIVALENCE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TEQUIVALENCE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TOR<TContextViewNode>::CalcSize(TMyDC&) {}
void  TOR<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TAND<TContextViewNode>::CalcSize(TMyDC&) {}
void  TAND<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TLESS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TLESS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TLESSEQ<TContextViewNode>::CalcSize(TMyDC&) {}
void  TLESSEQ<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TGREATER<TContextViewNode>::CalcSize(TMyDC&) {}
void  TGREATER<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TGREATEREQ<TContextViewNode>::CalcSize(TMyDC&) {}
void  TGREATEREQ<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TEQUALITY<TContextViewNode>::CalcSize(TMyDC&) {}
void  TEQUALITY<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TDIFFERENCE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TDIFFERENCE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TADDITION<TContextViewNode>::CalcSize(TMyDC&) {}
void  TADDITION<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TSUBTRACTION<TContextViewNode>::CalcSize(TMyDC&) {}
void  TSUBTRACTION<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TMULTIPLICATION<TContextViewNode>::CalcSize(TMyDC&) {}
void  TMULTIPLICATION<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TDIVISION<TContextViewNode>::CalcSize(TMyDC&) {}
void  TDIVISION<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TMODULO<TContextViewNode>::CalcSize(TMyDC&) {}
void  TMODULO<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TUNARYMINUS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TUNARYMINUS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TNOT<TContextViewNode>::CalcSize(TMyDC&) {}
void  TNOT<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TEXPRBLOCK<TContextViewNode>::CalcSize(TMyDC&) {}
void  TEXPRBLOCK<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPBODYSTATS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPBODYSTATS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPBODYPARSSTATS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPBODYPARSSTATS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPPARS<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TPPARS<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TPPARVALUE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TPPARVALUE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TPPARREF<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "! ", ": ", 0);
}
void  TPPARREF<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "! ", ": ", 0);
}

void  TPARGS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPARGS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPARGVALUE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPARGVALUE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TPARGREF<TContextViewNode>::CalcSize(TMyDC&) {}
void  TPARGREF<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TFBODY<TContextViewNode>::CalcSize(TMyDC&) {}
void  TFBODY<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TFPARS<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TFPARS<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TFPARVALUE<TContextViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TFPARVALUE<TContextViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TFARGS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TFARGS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void	TNUMCASE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TNUMCASE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void	TLABCASE<TContextViewNode>::CalcSize(TMyDC&) {}
void  TLABCASE<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TALTS1<TContextViewNode>::CalcSize(TMyDC&) {}
void  TALTS1<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TALTS2<TContextViewNode>::CalcSize(TMyDC&) {}
void  TALTS2<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TALT2<TContextViewNode>::CalcSize(TMyDC&) {}
void  TALT2<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TLABELS<TContextViewNode>::CalcSize(TMyDC&) {}
void  TLABELS<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TNUMLABEL<TContextViewNode>::CalcSize(TMyDC&) {}
void  TNUMLABEL<TContextViewNode>::Draw(TMyDC&, TPoint&) {}

void  TCHARLABEL<TContextViewNode>::CalcSize(TMyDC&) {}
void  TCHARLABEL<TContextViewNode>::Draw(TMyDC&, TPoint&) {}


typedef TScrolledTreeWindowView<TContextViewNode> TContextView;

DEFINE_RESPONSE_TABLE1(TContextView, TScrolledWindowView)
	EV_VN_MIRRORSUBTREE,
	EV_VN_NEWACTIVE,
END_RESPONSE_TABLE;

const char far* TScrolledTreeWindowView<TContextViewNode>::StaticName()
{
	return "Context View";
}

TScrolledTreeWindowView<TContextViewNode>::TScrolledTreeWindowView(TStructDocument& doc, TWindow *parent):
					TScrolledWindowView(doc, parent), Tree(doc.GetTree())
{
	StructDoc = &doc;
	ShiftDown = FALSE;
}

BOOL     TScrolledTreeWindowView<TContextViewNode>::VnNewActive(TNodeAddress *)
{
	AdjustScroller();
	Invalidate();
	return FALSE;
}


TSize    TScrolledTreeWindowView<TContextViewNode>::CalcSize()
{
	T   *root = Tree.GetRoot();

	if(root) {
		StructDoc->GetActiveAddress(TContextViewNode::ActiveAddress);
		root->CalcSize((TMyDC&)TClientDC(HWindow));
		return root->NodeSize();
	}
	return TSize(0, 0);
}

void TContextView::Paint(TDC& dc, BOOL, TRect&)
{
	dc.SetBkMode(TRANSPARENT);
	if(Tree.GetRoot()) {
		StructDoc->GetActiveAddress(TContextViewNode::ActiveAddress);
		Tree.GetRoot()->FlushNodePositions();
		Tree.GetRoot()->Draw((TMyDC&)dc, TPoint(0, 0));
	}
}



DEFINE_DOC_TEMPLATE_CLASS(TStructDocument, TContextView, TContextTemplate);
TContextTemplate contextTpl("Active Context View", "*.STR", 0, "STR",
									 dtAutoDelete | dtUpdateDir |dtAutoOpen);

