//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TTermEditor
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop
#include "mydc.h"
#include "tree.h"
#include "term.h"
#include "editor.h"



TTermEditor::TTermEditor()
{
	InsertHole(0, 0);
}

void     TTermEditor::InsertHole(TTerm *parent, int cn)
{
	Sort sort = parent ? parent->GetDesiredChildSort(cn) : TERMSORT;
#ifdef AUTOOP
	int	number = 0;
	Operator opfound;
	for(Operator op = OP_FIRST; number <=1 && op<=OP_LAST; op++) {
		if(GetOpSort(op) == sort) {
			opfound = op; number++;
		}
	}
	if(number == 1) {
		InsertOperator(parent, cn, opfound);
		return;
	}
#endif
	TTerm		*hole	= CreateOperatorNode((TTerm *)0,
		(Operator)sort);
	InsertChild(parent, cn, hole);
}

void		TTermEditor::InsertOperator(TTerm *parent, int cn, Operator op)
{
	TTerm		*newnode = CreateOperatorNode((TTerm *)0, op);
	InsertChild(parent, cn, newnode);
	int numchildholes = newnode->IsList() ? 1 : newnode->GetDesiredNumChildren();
	for(int i = 0; i<numchildholes; i++) {
		InsertHole(newnode, i);
	}
}


void     TTermEditor::Refine(TNodeAddress& address, Operator op)
{
	TTerm    *oldnode, *parent;
	int		 cn;

	oldnode = NodeAt(address);
	parent = GetParent(oldnode); cn = GetMyChildNumber(oldnode);
	RemoveChild(parent, cn);
	InsertOperator(parent, cn, op);
}

void     TTermEditor::Delete(TNodeAddress& address)
{
	TTerm    *oldnode, *newnode;

	oldnode = NodeAt(address);
	newnode = CreateOperatorNode((TTerm *)0,
				 (Operator)oldnode->GetSort());
	ReplaceSubtree(oldnode, newnode);
}



void     TTermEditor::ListInsert(TNodeAddress& address, int childnumber)
{
	TTerm    *listnode, *childnode;

	listnode  = NodeAt(address);
	childnode = CreateOperatorNode((TTerm *)0,
			(Operator)listnode->GetDesiredChildSort(childnumber));
	InsertChild(listnode, childnumber, childnode);
}


void     	TTermEditor::ListDelete(TNodeAddress& address, int childnumber)
{
	RemoveChild(NodeAt(address), childnumber);
}

BOOL        TTermEditor::ReadFromStream(istream& is)
{
	static TNodeAddress rootaddr;

	return ReadSubtreeFromStream(is, rootaddr);
}

BOOL        TTermEditor::ReadSubtreeFromStream(istream& is, TNodeAddress& addr)
{
	TTerm		*node = NodeAt(addr);
	TTerm 	*parent = GetParent(node);
	int 		 cn = GetMyChildNumber(node);

	RemoveChild(parent, cn);
	BOOL rv = ReadSubtreeFromStream(is, parent, cn);
	if(!rv) {
		InsertHole(parent, cn);
	}
	return rv;
}


BOOL        TTermEditor::ReadSubtreeFromStream(istream& is, TTerm *parent, int cn)
{
	Operator  	   op;
	char           opname[MAXOPNAMELEN+1], *p = opname;
	static char    opdata[MAXOPDATALEN+1];
	char           c = 0;
	int			   i = 0;
	BOOL           found = FALSE;
	TTerm         *node;
	BOOL           rv = FALSE;
	BOOL           childinserted = FALSE;

	while(is && isalnum(i = is.peek())) {
		is >> *p; p++;
	}

	*p = 0;
	for(op = OP_FIRSTHOLE; op<= OP_LAST; op++) {
		found = !stricmp(opname, GetOpName(op));
		if(found)
			break;
	}
	if(!found)
		goto quit;
	node = CreateOperatorNode((TTerm *)0, op);
	if(!node)
		goto quit;
	if(i == '{') {
		is >> c;
		p = opdata;
		while(is >> c && c!='}') {
			*p++ = c;
		}
		*p++ = 0;
		node->SetData(opdata, p-opdata);
		i = is.peek();
	}
	InsertChild(parent, cn, node);
	childinserted = TRUE;
	if(i == '[') {
		is >> c;
		do {
			if(!ReadSubtreeFromStream(is, node, node->GetNumChildren()))
				goto quit;
			c = 0;
		} while(is>>c && c==',');
		if(c!=']')
			goto quit;
	}
	if(!node->IsList() &&
		 node->GetNumChildren() != node->GetDesiredNumChildren() )
				goto quit;
	rv = TRUE;
quit:
	if(!rv && childinserted)
		RemoveChild(parent, cn);
	return rv;
}

