//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TFlowView
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop

//#pragma option -Jgx

#include <owl\menu.h>

#include "views.rc"
#include "tree.h"
#include "strucdoc.h"
#include "mydc.h"
#include "viewnode.h"
#include "viewtree.h"
#include "scrwview.h"
#include "scrtrwvw.h"


class   TFlowViewNode : public TViewNode {
public:
								 TFlowViewNode(int n = 0, int fixed = FALSE);
	inline TFlowViewNode *GetChild(int n) const;
	void                  CalcSizeFolded(TMyDC&);
	void                  DrawFolded(TMyDC&, TPoint& pos);
	void						 CalcSizeString(TMyDC&, const char const *);
	void						 DrawString(TMyDC&, TPoint&, const char const *);
	void						 CalcSizeHole(TMyDC&);
	void						 DrawHole(TMyDC&, TPoint&);
	BOOL						 DoPosFocusFolded(TMyDC& dc, TPoint& pos);
	void  					 CalcSizeDelimiters(TMyDC& dc, \
														 char *start, char *between, char *end);

	void  					 DrawDelimiters(TMyDC& dc, TPoint& pos, \
														 char *start, char *between, char *end);
	void						 DrawStatBox(TMyDC& dc, TPoint& pos);
	inline int            GetConnX() const;
	inline void           SetConnX(int connx);

	static TViewTree<TFlowViewNode>     *FlowTree;

private:
	int                   ConnX;
};

inline TFlowViewNode* TFlowViewNode::GetChild(int n) const
{  return (TFlowViewNode*)TTreeNode::GetChild(n);
}
inline int              TFlowViewNode::GetConnX() const
{  return ConnX; }
inline void             TFlowViewNode::SetConnX(int connx)
{  ConnX = connx; }



const int         DistX       = 6;
const int         DistY       = 8;
const int         TextBorderX = 6;
const int         TextBorderY = 2;
const int         BlockBorderX = 4;
const int         BlockBorderY = 4;
const int         HConnLen    = 6; //length of horiz connecting line
const int         VConnLen    = 6; //length of horiz connecting line
const int			BoxExtraX   = TextBorderX*2+RShadW;
const int			BoxExtraY	= TextBorderY*2+BShadH;

#define    LineColor    TColor(0,0,0)
#define    FocusColor   TColor(128, 128, 128)
#define    FillColor    TColor(255, 255, 255)
#define    ScopeColor   TColor(230, 230, 240)

static TPen			 ScopePen(ScopeColor, 1, PS_DOT);

static TPen        Pen(LineColor);
static TBrush      FillBrush(FillColor);
static TBrush      FocusBrush(FocusColor);
static TBrush      ShadowBrush(TColor::Black);

TViewTree<TFlowViewNode>	*TFlowViewNode::FlowTree = 0;

TFlowViewNode::TFlowViewNode(int n, int fixed):
	ConnX(0),TViewNode(n, fixed)
{
}

void				    TFlowViewNode::CalcSizeString(TMyDC& dc, const char const *s)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		SetNodeSize(dc.GetTextExtent(s, strlen(s)));
		SetConnX(NodeWidth()/2);
	}
}

void				    		 TFlowViewNode::DrawString(TMyDC& dc, TPoint& pos, const char const *s)
{
	if(DoPosFocusFolded(dc, pos)) {
		dc.TextOut(pos, s);
	}
}
BOOL						TFlowViewNode::DoPosFocusFolded(TMyDC& dc, TPoint& pos)
{
	TRect		bounding(pos, NodeSize());

	AddNodePosition(pos);
	if(!bounding.Touches(dc.GetClipBox()))
		return FALSE;
	FlushChildrenNodePositions();
	if(FlowTree->GetFocus() == this)
		dc.FillRect(bounding, FocusBrush);
	if(GetFolded()) {
		DrawFolded(dc, pos);
		return FALSE;
	}
	return TRUE;
}
void  					TFlowViewNode::CalcSizeFolded(TMyDC& dc)
{
	char	str[80];

	wsprintf(str, "%s <%s>", GetOpName(), GetSortName());
	TSize		sz = dc.GetTextExtent(str, strlen(str))+TSize(PlusSize+DistX, 0);
	if(GetSort() == SORT_STAT)
		sz += TSize(BoxExtraX, BoxExtraY);
	SetNodeSize(sz);
	SetConnX(sz.cx/2);
}


void  					TFlowViewNode::DrawFolded(TMyDC& dc, TPoint& p2)
{
	char	str[80];
	TPoint pos = p2;

	if(GetSort() == SORT_STAT) {
		DrawStatBox(dc, p2);
		pos = p2+TSize(TextBorderX, TextBorderY);
	}
	wsprintf(str, "%s <%s>", GetOpName(), GetSortName());
	dc.Plus(TPoint(pos.x, p2.y+(NodeSize().cy-PlusSize)/2));
	dc.TextOut(pos.x+PlusSize+DistX, pos.y, str);
}


void  TFlowViewNode::CalcSizeDelimiters(TMyDC& dc,
														 char *start, char *between, char *end)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int width       = startsize.cx;
		int height      = max(max(startsize.cy, betweensize.cy), endsize.cy);
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TFlowViewNode *child   = GetChild(i);
			child->CalcSize(dc);
			width += child->NodeWidth() +(i<numchildren-1?betweensize.cx:0);
			if(child->NodeHeight() > height)
				height = child->NodeHeight();
		}
		width += endsize.cx;
		if(GetSort() == SORT_STAT) {
			width += BoxExtraX; height += BoxExtraY;
		}
		SetNodeSize(TSize(width, height));
		SetConnX(width/2);
	}
}
void  TFlowViewNode::DrawDelimiters(TMyDC& dc, TPoint& pos,
														 char *start, char *between, char *end)
{	if(DoPosFocusFolded(dc, pos)) {
		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int numchildren = GetNumChildren();

		int x = pos.x;
		int y = pos.y;

		if(GetSort() == SORT_STAT) {
			DrawStatBox(dc, pos);
			x += TextBorderX; y += TextBorderY;
		}

		if(start)dc.TextOut(x, y, start);
		x += startsize.cx;
		for(int i = 0; i<numchildren; i++) {
			TFlowViewNode *child   = GetChild(i);
			child->Draw(dc, TPoint(x, y));
			x += child->NodeWidth();
			if(i<numchildren-1 && between) {
				dc.TextOut(x, y, between); x += betweensize.cx;
			}
		}
		if(end)dc.TextOut(x, y, end);
	}
}


void  TFlowViewNode::DrawStatBox(TMyDC& dc, TPoint& pos)
{
	TRect     rect(pos, NodeSize()-TSize(RShadW, BShadH));
	dc.ShadowRectangle(rect, FlowTree->GetFocus() == this ? FocusBrush : FillBrush,
							 Pen, ShadowBrush);
}



void  TSTATHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	SetNodeSize(dc.GetTextExtent("? <STAT>", 8)+TSize(BoxExtraX, BoxExtraY));
	SetConnX(NodeWidth()/2);
}

void  TSTATHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		DrawStatBox(dc, pos);
		dc.TextOut(pos.x+TextBorderX, pos.y+TextBorderY, "? <STAT>");
	}
}


void						TFlowViewNode::CalcSizeHole(TMyDC& dc)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	CalcSizeString(dc, s);
}

void						TFlowViewNode::DrawHole(TMyDC& dc, TPoint& pos)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	DrawString(dc, pos, s);
}


void  TPROGHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPROGHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TDECHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TDECSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TNDECHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TNDECHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TIDHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TIDHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TTYPEHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TTYPEHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
/*void  TSTATHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }*/
void  TSTATSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TVARSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TVARSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS1HOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS1HOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS2HOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS2HOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALT2HOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALT2HOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPBODYHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPBODYHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFBODYHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFBODYHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFARGSHOLE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFARGSHOLE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }


void  TPROG<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();  stats->CalcSize(dc);

		SetNodeSize(TSize(2*BlockBorderX+max(decs->NodeSize().cx, stats->NodeSize().cx),
								2*BlockBorderY+decs->NodeSize().cy+ArrowLength+stats->NodeSize().cy));
		SetConnX(NodeWidth()/2);
	}
}


void  TPROG<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int			y = pos.y+BlockBorderY;

		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();

		dc.SelectObject(ScopePen);
		dc.RectangleOutline(TRect(pos, NodeSize()));
		dc.RestorePen();

		decs->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += decs->NodeHeight()+ArrowLength;

		dc.Arrow(TPoint(pos.x+BlockBorderX+stats->GetConnX(), y), TMyDC::DOWN);
		stats->Draw(dc, TPoint(pos.x+BlockBorderX, y));
	}
}


void  TDECS<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    semicolonsize;
		dc.GetTextExtent(";",     1, semicolonsize);

		int width       = 0;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortDEC<TFlowViewNode> *dec   = GetDeclaration(i);
			dec->CalcSize(dc);
			int sw = dec->NodeSize().cx+semicolonsize.cx;
			if(sw>width) width = sw;
			height += dec->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}
void  TDECS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int y = pos.y;
		int numchildren = GetNumChildren();
		TSize    semicolonsize;
		dc.GetTextExtent(";",     1, semicolonsize);

		for(int i = 0; i<numchildren; i++) {
			TSortDEC<TFlowViewNode> *dec   = GetDeclaration(i);
			if(i) dc.TextOut(pos.x, y, ";");
			dec->Draw(dc, TPoint(pos.x+semicolonsize.cx, y));
			y += dec->NodeSize().cy;
		}
	}
}


void  TNDECS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "var  ", ", ", 0);
}
void  TNDECS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "var  ", ", ", 0);
}
void  TNDEC<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc,  0, ":", 0);
}
void  TNDEC<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ":", 0);
}

void  TPROCDEC<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TFlowViewNode>	 *id = GetProcName();
		id->CalcSize(dc);
		TSortPBODY<TFlowViewNode> *pbody = GetBody();
		pbody->CalcSize(dc);

		TSize    procsize, eqsize;
		dc.GetTextExtent("proc ", 5, procsize);
		dc.GetTextExtent(" = ",   3, eqsize);

		SetNodeSize(TSize(procsize.cx + id->NodeWidth() + eqsize.cx + pbody->NodeWidth(),
								pbody->NodeHeight()));
	}
}

void  TPROCDEC<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		TSize    procsize, eqsize;
		dc.GetTextExtent("proc ", 5, procsize);
		dc.GetTextExtent(" = ",   3, eqsize);
		dc.TextOut(pos, "proc");
		int	x = pos.x+procsize.cx;
		TSortID<TFlowViewNode>	 *id = GetProcName();
		id->Draw(dc, TPoint(x, pos.y));
		x += id->NodeWidth(); dc.TextOut(x, pos.y, " = ");
		x += eqsize.cx;
		TSortPBODY<TFlowViewNode> *pbody = GetBody();
		pbody->Draw(dc, TPoint(x, pos.y));
	}
}


void  TFUNCDEC<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TFlowViewNode>	 *id = GetFuncName();
		id->CalcSize(dc);
		TSortFBODY<TFlowViewNode> *fbody = GetBody();
		fbody->CalcSize(dc);

		TSize    FUNCsize, eqsize;
		dc.GetTextExtent("func ", 5, FUNCsize);
		dc.GetTextExtent(" = ",   3, eqsize);

		SetNodeSize(TSize(FUNCsize.cx + id->NodeWidth() + eqsize.cx + fbody->NodeWidth(),
								fbody->NodeHeight()));
	}
}

void  TFUNCDEC<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	if(DoPosFocusFolded(dc, pos)) {
		TSize    FUNCsize, eqsize;
		dc.GetTextExtent("func ", 5, FUNCsize);
		dc.GetTextExtent(" = ",   3, eqsize);
		dc.TextOut(pos, "func");
		int	x = pos.x+FUNCsize.cx;
		TSortID<TFlowViewNode>	 *id = GetFuncName();
		id->Draw(dc, TPoint(x, pos.y));
		x += id->NodeWidth(); dc.TextOut(x, pos.y, " = ");
		x += eqsize.cx;
		TSortFBODY<TFlowViewNode> *fbody = GetBody();
		fbody->Draw(dc, TPoint(x, pos.y));
	}
}


void  TID<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TID<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }


void  TINTTYPE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "int"); }
void  TINTTYPE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "int"); }
void  TCHARTYPE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "char"); }
void  TCHARTYPE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "char"); }
void  TREALTYPE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "real"); }
void  TREALTYPE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "real"); }
void  TBOOLTYPE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "bool"); }
void  TBOOLTYPE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "bool"); }

void  TSTATS<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortSTAT<TFlowViewNode>  *stat;
		int   numchildren = GetNumChildren();
		int   offset;
		int   width = 0, height = 0;
		int   connx = 0;

		for(int i = 0; i<numchildren; i++) {
			stat   = GetStatement(i);
			stat->CalcSize(dc);
			offset = connx-stat->GetConnX();
			if(offset < 0) {
				connx     -= offset;
				width     -= offset;
				offset     = 0;
			}
			if(offset+stat->NodeWidth() > width)
				width = offset+stat->NodeWidth();
			height += stat->NodeHeight() + (i?DistY:0);
		}
		SetConnX(connx);
		SetNodeSize(TSize(width, height));
	}
}

void  TSTATS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(DoPosFocusFolded(dc, pos)) {
		TSize    semicolonsize;
		dc.GetTextExtent(";", 1, semicolonsize);
		int connx = GetConnX();
		int y = pos.y;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TFlowViewNode>  *stat = GetStatement(i);

			if(i) {
				dc.MoveTo(pos.x+connx, y-DistY); dc.LineTo(pos.x+connx, y);
				dc.Arrow(TPoint(pos.x+connx, y), TMyDC::DOWN);
			}
			stat->Draw(dc, TPoint(pos.x+connx - stat->GetConnX(), y));
			y += stat->NodeHeight()+DistY;
		}
	}
}

void  TSKIP<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "skip", 0, 0);
}
void  TSKIP<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "skip", 0, 0);
}

void  TCONCASSIGN<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, " := ", 0);
}
void  TCONCASSIGN<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, " := ", 0);
}


void  TPROCCALL<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, 0, 0);
}
void  TPROCCALL<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, 0, 0);
}

void  TPROCCALLARGS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, "(", ")");
}
void  TPROCCALLARGS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, "(", ")");
}

void  TIFTHEN<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TFlowViewNode>  *guard = GetGuard(); guard->CalcSize(dc);
		TSortSTATS<TFlowViewNode> *stats = GetTrueStatements(); stats->CalcSize(dc);

		TSize gsize = guard->NodeSize() + TSize(guard->NodeHeight()+TextBorderY*2, TextBorderY*2);

		SetNodeSize(TSize(max(stats->GetConnX() + HConnLen*2 + gsize.cx ,
									stats->NodeWidth() + DistX),
								gsize.cy+DistY+stats->NodeHeight()+VConnLen));
		SetConnX(max(NodeWidth()/2, stats->GetConnX()+HConnLen+gsize.cx/2));
	}
}

void  TIFTHEN<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	if(DoPosFocusFolded(dc, pos)) {
		TSortEXPR<TFlowViewNode>  *guard = GetGuard();
		TSortSTATS<TFlowViewNode> *stats = GetTrueStatements();
		int                     	d, d2, gx, rx;
		TSize                   	gsize;

		d2    = guard->NodeHeight()+TextBorderY*2; d = d2/2;
		gsize = guard->NodeSize() + TSize(d2, TextBorderY*2);
		gx    = pos.x + GetConnX() - gsize.cx/2;


		dc.ShadowRectangle(TRect(TPoint(gx, pos.y),gsize), FillBrush, Pen, ShadowBrush);
		dc.MoveTo(gx+d, pos.y); dc.LineTo(gx, pos.y+d); dc.LineTo(gx+d, pos.y+d2);
		rx = gx+gsize.cx-1;
		dc.MoveTo(rx-d, pos.y); dc.LineTo(rx, pos.y+d); dc.LineTo(rx-d, pos.y+d2);

		dc.MoveTo(gx, pos.y+d); dc.LineTo(pos.x+stats->GetConnX(), pos.y+d);
		dc.LineTo(pos.x+stats->GetConnX(), pos.y+gsize.cy+DistY);
		dc.Arrow(TPoint(pos.x+stats->GetConnX(), pos.y+gsize.cy+DistY), TMyDC::DOWN);


		dc.MoveTo(rx, pos.y+d); dc.LineTo(pos.x+NodeWidth()-1, pos.y+d);
		dc.LineTo(pos.x+NodeWidth()-1, pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+stats->GetConnX(), pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+stats->GetConnX(), pos.y+NodeHeight()-VConnLen);
		guard->Draw(dc, TPoint(gx+d, pos.y+TextBorderY));
		stats->Draw(dc, TPoint(pos.x, pos.y+gsize.cy+DistY));
	}
}

void  TIFTHENELSE<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TFlowViewNode>  *guard  = GetGuard();            guard->CalcSize(dc);
		TSortSTATS<TFlowViewNode> *tstats = GetTrueStatements();  tstats->CalcSize(dc);
		TSortSTATS<TFlowViewNode> *fstats = GetFalseStatements(); fstats->CalcSize(dc);

		TSize gsize = guard->NodeSize() + TSize(guard->NodeHeight()+TextBorderY*2, TextBorderY*2);

		int		t1, t2, f1, f2, d;

		t1 = tstats->GetConnX();  t2 = tstats->NodeWidth()-t1;
		f1 = fstats->GetConnX();  f2 = fstats->NodeWidth()-f1;

		d = gsize.cx+HConnLen*2-(t2+DistX+f1);
		if(d<0) d = 0;
		SetNodeSize(TSize(t1+t2 + DistX + d + f1+f2,
								gsize.cy+DistY+max(tstats->NodeHeight(),fstats->NodeHeight())+VConnLen));
		SetConnX(t1+(t2+d+DistX+f1)/2);
	}
}

void  TIFTHENELSE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	if(DoPosFocusFolded(dc, pos)) {
		TSortEXPR<TFlowViewNode>  *guard = GetGuard();
		TSortSTATS<TFlowViewNode> *tstats = GetTrueStatements();
		TSortSTATS<TFlowViewNode> *fstats = GetFalseStatements();
		int                     	d, d2, gx, rx, fx;
		TSize                   	gsize;

		d2    = guard->NodeHeight()+TextBorderY*2; d = d2/2;
		gsize = guard->NodeSize() + TSize(d2, TextBorderY*2);
		gx    = pos.x + GetConnX() - gsize.cx/2;
		fx		= pos.x+NodeWidth()-fstats->NodeWidth();

		/* Draw Guard */
		dc.ShadowRectangle(TRect(TPoint(gx, pos.y),gsize), FillBrush, Pen, ShadowBrush);
		dc.MoveTo(gx+d, pos.y); dc.LineTo(gx, pos.y+d); dc.LineTo(gx+d, pos.y+d2);
		rx = gx+gsize.cx-1;
		dc.MoveTo(rx-d, pos.y); dc.LineTo(rx, pos.y+d); dc.LineTo(rx-d, pos.y+d2);

		/* Draw TRUE line */
		dc.MoveTo(gx, pos.y+d); dc.LineTo(pos.x+tstats->GetConnX(), pos.y+d);
		dc.LineTo(pos.x+tstats->GetConnX(), pos.y+gsize.cy+DistY);
		dc.Arrow(TPoint(pos.x+tstats->GetConnX(), pos.y+gsize.cy+DistY), TMyDC::DOWN);

		/* Draw FALSE line */
		dc.MoveTo(rx, pos.y+d); dc.LineTo(fx+fstats->GetConnX(), pos.y+d);
		dc.LineTo(fx+fstats->GetConnX(), pos.y+gsize.cy+DistY);
		dc.Arrow(TPoint(fx+fstats->GetConnX(), pos.y+gsize.cy+DistY), TMyDC::DOWN);

		dc.MoveTo(fx+fstats->GetConnX(), pos.y+gsize.cy+DistY+fstats->NodeHeight());
		dc.LineTo(fx+fstats->GetConnX(), pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+tstats->GetConnX(), pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+tstats->GetConnX(), pos.y+gsize.cy+DistY+tstats->NodeHeight());

		guard->Draw(dc, TPoint(gx+d,  pos.y+TextBorderY));
		tstats->Draw(dc, TPoint(pos.x, pos.y+gsize.cy+DistY));
		fstats->Draw(dc, TPoint(fx,    pos.y+gsize.cy+DistY));
	}
}

void  TWHILE<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TFlowViewNode>  *guard = GetGuard(); guard->CalcSize(dc);
		TSortSTATS<TFlowViewNode> *stats = GetStatements(); stats->CalcSize(dc);
		TSize                   gsize;

		gsize = guard->NodeSize() + TSize(guard->NodeHeight()+TextBorderY*2, TextBorderY*2);

		SetNodeSize(TSize( max(DistX + stats->GetConnX() + HConnLen*2 + gsize.cx ,
									  stats->NodeWidth() + DistX*2),
								 ArrowWidth/2+VConnLen+gsize.cy+DistY+stats->NodeHeight()+VConnLen));
		SetConnX(max(NodeWidth()/2, DistX+stats->GetConnX()+HConnLen+gsize.cx/2));
	}
}


void  TWHILE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{
	if(DoPosFocusFolded(dc, pos)) {
		int                        d, d2, gx, gy, rx;
		TSize                      gsize;
		TSortEXPR<TFlowViewNode>  *guard = GetGuard();
		TSortSTATS<TFlowViewNode> *stats = GetStatements();

		d2    = guard->NodeHeight()+TextBorderY*2; d = d2/2;
		gsize = guard->NodeSize() + TSize(d2, TextBorderY*2);
		gx    = pos.x + GetConnX() - gsize.cx/2; gy = pos.y+VConnLen+ArrowWidth/2;

		dc.MoveTo(pos.x+GetConnX(), pos.y); dc.LineTo(pos.x+GetConnX(), gy);

		/* draw guard surroundings*/
		dc.ShadowRectangle(TRect(TPoint(gx, gy),gsize), FillBrush, Pen, ShadowBrush);
		dc.MoveTo(gx+d, gy); dc.LineTo(gx, gy+d); dc.LineTo(gx+d, gy+d2);
		rx = gx+gsize.cx-1;
		dc.MoveTo(rx-d, gy); dc.LineTo(rx, gy+d); dc.LineTo(rx-d, gy+d2);

		/* draw TRUE line to stats */
		dc.MoveTo(gx, gy+d); dc.LineTo(pos.x+DistX+stats->GetConnX(), gy+d);
		dc.LineTo(pos.x+DistX+stats->GetConnX(), gy+gsize.cy+DistY);
		dc.Arrow(TPoint(pos.x+DistX+stats->GetConnX(), gy+gsize.cy+DistY), TMyDC::DOWN);

		/* from stats back to top */
		dc.MoveTo(pos.x+DistX+stats->GetConnX(), pos.y+NodeHeight()-VConnLen);
		dc.LineTo(pos.x+DistX+stats->GetConnX(), pos.y+NodeHeight()-1);
		dc.LineTo(pos.x,       pos.y+NodeHeight()-1);
		dc.LineTo(pos.x,       pos.y+ArrowWidth/2);
		dc.LineTo(pos.x+GetConnX(), pos.y+ArrowWidth/2);
		dc.Arrow(TPoint(pos.x+DistX+stats->GetConnX()+ArrowLength/2,
							  pos.y+ArrowWidth/2), TMyDC::RIGHT);
		/* draw FALSE line to bottom */
		dc.MoveTo(rx, gy+d);
		dc.LineTo(pos.x+NodeWidth()-1, gy+d);
		dc.LineTo(pos.x+NodeWidth()-1, pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+GetConnX(),  pos.y+NodeHeight()-1);

		/* Guard itself */
		guard->Draw(dc, TPoint(gx+d, gy+TextBorderY));
		/* Stat itself */
		stats->Draw(dc, TPoint(pos.x+DistX, gy+gsize.cy+DistY));
	}
}


void  TREAD<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "read(", ",", ")");
}
void  TREAD<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "read(", ",", ")");
}

void  TWRITE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "write(", ",", ")");
}
void  TWRITE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "write(", ",", ")");
}

void  TFOR<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TFlowViewNode>    *id  	= GetVariable(); 			id->CalcSize(dc);
		TSortEXPR<TFlowViewNode>  *from 	= GetFromExpression(); 	from->CalcSize(dc);
		TSortEXPR<TFlowViewNode>  *by  	= GetByExpression(); 	by->CalcSize(dc);
		TSortEXPR<TFlowViewNode>  *to  	= GetToExpression(); 	to->CalcSize(dc);
		TSortSTATS<TFlowViewNode> *stats = GetStatements(); 		stats->CalcSize(dc);
		int		maxh, statconnx, connx, owidth, statwidth;
		TSize    assignsize, testsize, incrsize, plussize;

		dc.GetTextExtent(" := ",   4, assignsize);
		assignsize.cx += id->NodeWidth()+from->NodeWidth()+BoxExtraX;
		maxh = max(id->NodeHeight(), from->NodeHeight());
		if(maxh>assignsize.cy)
			assignsize.cy = maxh;
		assignsize.cy += BoxExtraY;

//		gsize = guard->NodeSize() + TSize(guard->NodeHeight()+TextBorderY*2, TextBorderY*2);
		dc.GetTextExtent("<=",   2, testsize);
		testsize.cx += id->NodeWidth()+to->NodeWidth();
		maxh = max(id->NodeHeight(), to->NodeHeight());
		if(maxh>testsize.cy)
			testsize.cy = maxh;
		testsize += TSize(testsize.cy+TextBorderY*2+RShadW, TextBorderY*2+BShadH);

		dc.GetTextExtent(" := ", 4, incrsize);
		dc.GetTextExtent("+",    1, plussize);
		incrsize.cx += 2*id->NodeWidth()+by->NodeWidth()+plussize.cx+BoxExtraX;
		maxh = max(id->NodeHeight(), by->NodeHeight());
		if(maxh>incrsize.cy)
			incrsize.cy = maxh;
		incrsize.cy += BoxExtraY;


		statconnx = DistX+stats->GetConnX();
		if(statconnx < DistX+incrsize.cx/2)
			statconnx = DistX+incrsize.cx/2;
		statwidth = statconnx+max(stats->NodeWidth()-stats->GetConnX(),
										  incrsize.cx/2);

		connx = statconnx+HConnLen+testsize.cx/2;

		if(connx < assignsize.cx/2)
			connx = assignsize.cx/2;
		owidth = max(connx+assignsize.cx/2, connx+testsize.cx/2+HConnLen);



		SetNodeSize(TSize(max(statwidth+DistX, owidth),
								assignsize.cy + testsize.cy +
								stats->NodeHeight() + incrsize.cy + DistY*3 + VConnLen));
		SetConnX(connx);
	}
}

void  TFOR<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(DoPosFocusFolded(dc, pos)) {
		TSortID<TFlowViewNode>  	*id  	 = GetVariable();
		TSortEXPR<TFlowViewNode>  	*from  = GetFromExpression();
		TSortEXPR<TFlowViewNode>  	*by  	 = GetByExpression();
		TSortEXPR<TFlowViewNode>  	*to  	 = GetToExpression();
		TSortSTATS<TFlowViewNode> 	*stats = GetStatements();

		int		maxh, x, y, idx, idy, statconnx, gx, gy, d, d2, rx, ix, iy;
		TSize    assignsize, testsize, incrsize, plussize, gsize;

		dc.GetTextExtent(" := ",   4, assignsize);
		assignsize.cx += id->NodeWidth()+from->NodeWidth()+BoxExtraX;
		maxh = max(id->NodeHeight(), from->NodeHeight());
		if(maxh>assignsize.cy)
			assignsize.cy = maxh;
		assignsize.cy += BoxExtraY;

		x = pos.x+GetConnX()-assignsize.cx/2;
		dc.ShadowRectangle(TRect(TPoint(x, pos.y),assignsize-TSize(RShadW, BShadH)), FillBrush, Pen, ShadowBrush);
		idx = x+TextBorderX;
		idy = pos.y+TextBorderY;
		x = idx+id->NodeWidth();
		dc.TextOut(x, idy, " := "); x+=dc.GetTextExtent(" := ", 4).cx;
		from->Draw(dc, TPoint(x, idy));

		dc.MoveTo(pos.x+GetConnX(), pos.y+assignsize.cy);

//		gsize = guard->NodeSize() + TSize(guard->NodeHeight()+TextBorderY*2, TextBorderY*2);
		dc.GetTextExtent("<=",   2, gsize);
		gsize.cx += id->NodeWidth()+to->NodeWidth();
		maxh = max(id->NodeHeight(), to->NodeHeight());
		if(maxh>gsize.cy)
			gsize.cy = maxh;
		gsize += TSize(gsize.cy+TextBorderY*2, TextBorderY*2);

		d2    = gsize.cy; d = d2/2;
		gx    = pos.x + GetConnX() - (gsize.cx+RShadW)/2;
		gy 	= pos.y+assignsize.cy+DistY;
		dc.LineTo(pos.x+GetConnX(), gy);

		dc.ShadowRectangle(TRect(TPoint(gx, gy),gsize), FillBrush, Pen, ShadowBrush);
		dc.MoveTo(gx+d, gy); dc.LineTo(gx, gy+d); dc.LineTo(gx+d, gy+d2);
		rx = gx+gsize.cx-1;
		dc.MoveTo(rx-d, gy); dc.LineTo(rx, gy+d); dc.LineTo(rx-d, gy+d2);

		x = gx+TextBorderX;
		y = gy+TextBorderY;
		id->Draw(dc, TPoint(x, y)); x += id->NodeWidth();
		dc.TextOut(x, y, "<="); x += dc.GetTextExtent("<=", 2).cx;

		to->Draw(dc, TPoint(x, y));

		dc.GetTextExtent(" := ", 4, incrsize);
		dc.GetTextExtent("+",    1, plussize);
		incrsize.cx += 2*id->NodeWidth()+by->NodeWidth()+plussize.cx+BoxExtraX;
		maxh = max(id->NodeHeight(), by->NodeHeight());
		if(maxh>incrsize.cy)
			incrsize.cy = maxh;
		incrsize.cy += BoxExtraY;

		statconnx = pos.x+DistX+stats->GetConnX();
		if(statconnx < pos.x+DistX+incrsize.cx/2)
			statconnx = pos.x+DistX+incrsize.cx/2;


		/* draw TRUE line to stats */
		dc.MoveTo(gx, gy+d); dc.LineTo(statconnx, gy+d);
		y = gy+gsize.cy+BShadH+DistY;
		dc.LineTo(statconnx, y);
		dc.Arrow(TPoint(statconnx, y), TMyDC::DOWN);

		stats->Draw(dc, TPoint(statconnx-stats->GetConnX(), y));
		y+= stats->NodeHeight();

		dc.MoveTo(statconnx, y);
		y+= DistY;
		dc.LineTo(statconnx, y);
		dc.Arrow(TPoint(statconnx, y), TMyDC::DOWN);

		ix = statconnx-incrsize.cx/2; iy = y;
		dc.ShadowRectangle(TRect(TPoint(ix, y),incrsize-TSize(RShadW, BShadH)), FillBrush, Pen, ShadowBrush);
		x = ix+TextBorderX; y+= TextBorderY;
		id->Draw(dc, TPoint(x, y)); x += id->NodeWidth();
		dc.TextOut(x, y, " := ");   x += dc.GetTextExtent(" := ", 4).cx;
		id->Draw(dc, TPoint(x, y)); x += id->NodeWidth();
		dc.TextOut(x, y, "+");      x += plussize.cx;
		by->Draw(dc, TPoint(x, y));
		y = iy+incrsize.cy;
		/* from stats back to top */
		dc.MoveTo(statconnx,        y);
		dc.LineTo(statconnx,        pos.y+NodeHeight()-1);
		dc.LineTo(pos.x,            pos.y+NodeHeight()-1);
		dc.LineTo(pos.x,            pos.y+assignsize.cy+DistY/2);
		dc.LineTo(pos.x+GetConnX(), pos.y+assignsize.cy+DistY/2);
		dc.Arrow(TPoint(statconnx+ArrowLength/2,
							 pos.y+assignsize.cy+DistY/2), TMyDC::RIGHT);
		/* draw FALSE line to bottom */
		dc.MoveTo(rx, gy+d);
		dc.LineTo(pos.x+NodeWidth()-1, gy+d);
		dc.LineTo(pos.x+NodeWidth()-1, pos.y+NodeHeight()-1);
		dc.LineTo(pos.x+GetConnX(),  pos.y+NodeHeight()-1);

		id->Draw(dc, TPoint(idx, idy));
	}
}

void  TBLOCK<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TFlowViewNode>*  stats = GetStatements(); stats->CalcSize(dc);

		SetNodeSize(TSize(2*BlockBorderX+max(decs->NodeSize().cx, stats->NodeSize().cx),
								2*BlockBorderY+decs->NodeSize().cy+ArrowLength+stats->NodeSize().cy));
		SetConnX(NodeWidth()/2);
	}
}


void  TBLOCK<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int			y = pos.y+BlockBorderY;

		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();

		dc.SelectObject(ScopePen);
		dc.RectangleOutline(TRect(pos,NodeSize()));
		dc.RestorePen();

		decs->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += decs->NodeHeight()+ArrowLength;

		dc.Arrow(TPoint(pos.x+BlockBorderX+stats->GetConnX(), y), TMyDC::DOWN);
		stats->Draw(dc, TPoint(pos.x+BlockBorderX, y));
	}
}



void  TVARS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TVARS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TEXPRS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc,  0, ", ", 0);
}
void  TEXPRS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TVARVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		GetVariable()->CalcSize(dc); SetNodeSize(GetVariable()->NodeSize());
	}
}

void  TVARVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos))
		GetVariable()->Draw(dc, pos);
}
void  TFUNCCALL<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, 0, 0);
}
void  TFUNCCALL<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, 0, 0);
}

void  TFUNCCALLARGS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, "(", ")");
}
void  TFUNCCALLARGS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, "(", ")");
}

void  TNUMVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TNUMVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TCHARVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TCHARVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TIMPLICATION<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " => ", ")");
}
void  TIMPLICATION<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " => ", ")");
}
void  TEQUIVALENCE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " == ", ")");
}
void  TEQUIVALENCE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " == ", ")");
}

void  TOR<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " \\/ ", ")");
}
void  TOR<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " \\/ ", ")");
}

void  TAND<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " /\\ ", ")");
}
void  TAND<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " /\\ ", ")");
}

void  TLESS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " < ", ")");
}
void  TLESS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " < ", ")");
}


void  TLESSEQ<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " <= ", ")");
}
void  TLESSEQ<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " <= ", ")");
}


void  TGREATER<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " > ", ")");
}
void  TGREATER<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " > ", ")");
}


void  TGREATEREQ<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " >= ", ")");
}
void  TGREATEREQ<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " >= ", ")");
}


void  TEQUALITY<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " == ", ")");
}
void  TEQUALITY<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " == ", ")");
}

void  TDIFFERENCE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " /= ", ")");
}
void  TDIFFERENCE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " /= ", ")");
}

void  TADDITION<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " + ", ")");
}
void  TADDITION<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " + ", ")");
}

void  TSUBTRACTION<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " - ", ")");
}
void  TSUBTRACTION<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " - ", ")");
}

void  TMULTIPLICATION<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", "*", ")");
}
void  TMULTIPLICATION<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", "*", ")");
}

void  TDIVISION<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", "/", ")");
}
void  TDIVISION<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", "/", ")");
}

void  TMODULO<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " mod ", ")");
}
void  TMODULO<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " mod ", ")");
}


void  TUNARYMINUS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "-", 0, 0);
}
void  TUNARYMINUS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "-", 0, 0);
}

void  TNOT<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "NOT ", 0, 0);
}
void  TNOT<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "NOT ", 0, 0);
}
void  TEXPRBLOCK<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();  stats->CalcSize(dc);
		TSortEXPR<TFlowViewNode>*   expr  = GetExpression();   expr->CalcSize(dc);

		SetNodeSize(TSize(2*BlockBorderX+max(max(decs->NodeSize().cx, stats->NodeSize().cx),expr->NodeWidth()),
								2*BlockBorderY+decs->NodeHeight()+ArrowLength+stats->NodeHeight()+DistY+expr->NodeHeight()));
		SetConnX(NodeWidth()/2);
	}
}


void  TEXPRBLOCK<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int			y = pos.y+BlockBorderY;

		TSortDECS<TFlowViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();
		TSortEXPR<TFlowViewNode>*   expr  = GetExpression();

		dc.SelectObject(ScopePen);
		dc.RectangleOutline(TRect(pos, NodeSize()));
		dc.RestorePen();

		decs->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += decs->NodeHeight()+ArrowLength;

		dc.Arrow(TPoint(pos.x+BlockBorderX+stats->GetConnX(), y), TMyDC::DOWN);
		stats->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += stats->NodeHeight()+DistY;
		expr->Draw(dc, TPoint(pos.x+BlockBorderX, y));
	}
}


void  TPBODYSTATS<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortSTAT<TFlowViewNode>  *stat;
		int   numchildren = GetNumChildren();
		int   offset;
		int   width = 0, height = 0;
		int   connx = 0;

		for(int i = 0; i<numchildren; i++) {
			stat   = GetStatement(i);
			stat->CalcSize(dc);
			offset = connx-stat->GetConnX();
			if(offset < 0) {
				connx     -= offset;
				width     -= offset;
				offset     = 0;
			}
			if(offset+stat->NodeWidth() > width)
				width = offset+stat->NodeWidth();
			height += stat->NodeHeight() + (i?DistY:0);
		}
		SetConnX(connx);
		SetNodeSize(TSize(width, height));
	}
}

void  TPBODYSTATS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(DoPosFocusFolded(dc, pos)) {
		int connx = GetConnX();
		int y = pos.y;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TFlowViewNode>  *stat = GetStatement(i);

			if(i) {
				dc.MoveTo(pos.x+connx, y-DistY); dc.LineTo(pos.x+connx, y);
				dc.Arrow(TPoint(pos.x+connx, y), TMyDC::DOWN);
			}
			stat->Draw(dc, TPoint(pos.x+connx - stat->GetConnX(), y));
			y += stat->NodeHeight()+DistY;
		}
	}
}

void  TPBODYPARSSTATS<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortPPARS<TFlowViewNode>*  pars  = GetProcParameters(); pars->CalcSize(dc);
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();  stats->CalcSize(dc);

		SetNodeSize(TSize(2*BlockBorderX+max(pars->NodeSize().cx, stats->NodeSize().cx),
								2*BlockBorderY+pars->NodeSize().cy+ArrowLength+stats->NodeHeight()));
		SetConnX(NodeWidth()/2);
	}
}


void  TPBODYPARSSTATS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int			y = pos.y+BlockBorderY;
		TSortPPARS<TFlowViewNode>*  pars  = GetProcParameters();
		TSortSTATS<TFlowViewNode>*  stats = GetStatements();

		dc.SelectObject(ScopePen);
		dc.RectangleOutline(TRect(pos, NodeSize()));
		dc.RestorePen();

		pars->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += pars->NodeHeight()+ArrowLength;

		dc.Arrow(TPoint(pos.x+BlockBorderX+stats->GetConnX(), y), TMyDC::DOWN);
		stats->Draw(dc, TPoint(pos.x+BlockBorderX, y));
	}
}


void  TPPARS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TPPARS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}


void  TPPARVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TPPARVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TPPARREF<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "! ", ": ", 0);
}
void  TPPARREF<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "! ", ": ", 0);
}

void  TPARGS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TPARGS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TPARGVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TPARGVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TPARGREF<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "! ", ": ", 0);
}
void  TPARGREF<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "! ", ": ", 0);
}


void  TFBODY<TFlowViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortFPARS<TFlowViewNode>*  pars  = GetFuncParameters(); pars->CalcSize(dc);
		TSortEXPR<TFlowViewNode>*   expr  = GetExpression(); 		expr->CalcSize(dc);

		SetNodeSize(TSize(2*BlockBorderX+max(pars->NodeSize().cx, expr->NodeSize().cx),
								2*BlockBorderY+pars->NodeSize().cy+DistY+expr->NodeHeight()));
		SetConnX(NodeWidth()/2);
	}
}


void  TFBODY<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	if(DoPosFocusFolded(dc, pos)) {
		int			y = pos.y+BlockBorderY;
		TSortFPARS<TFlowViewNode>*  pars  = GetFuncParameters();
		TSortEXPR<TFlowViewNode>*   expr  = GetExpression();

		dc.SelectObject(ScopePen);
		dc.RectangleOutline(TRect(pos, NodeSize()));
		dc.RestorePen();

		pars->Draw(dc, TPoint(pos.x+BlockBorderX, y));
		y += pars->NodeHeight()+DistY;

		expr->Draw(dc, TPoint(pos.x+BlockBorderX, y));
	}
}






void  TFPARS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TFPARS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TFPARVALUE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TFPARVALUE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TFARGS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ", ? ", 0);
}
void  TFARGS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ", ? ", 0);
}

void	TNUMCASE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "NUMCASE -- flow diagram view not available");
}

void  TNUMCASE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "NUMCASE -- flow diagram view not available");
}

void	TLABCASE<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "LABCASE -- flow diagram view not available");
}

void  TLABCASE<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "LABCASE -- flow diagram view not available");
}

void  TALTS1<TFlowViewNode>::CalcSize(TMyDC&)
{
}

void  TALTS1<TFlowViewNode>::Draw(TMyDC&, TPoint&)
{
}

void  TALTS2<TFlowViewNode>::CalcSize(TMyDC&)
{
}

void  TALTS2<TFlowViewNode>::Draw(TMyDC&, TPoint&)
{
}
void  TALT2<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ": ", 0);
}
void  TALT2<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ": ", 0);
}
void  TLABELS<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ",", 0);
}
void  TLABELS<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ",", 0);
}

void  TNUMLABEL<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TNUMLABEL<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TCHARLABEL<TFlowViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TCHARLABEL<TFlowViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

typedef TScrolledTreeWindowView<TFlowViewNode> TFlowView;

DEFINE_RESPONSE_TABLE1(TFlowView, TScrolledWindowView)
	EV_WM_CHAR,
	EV_WM_LBUTTONDOWN,
	EV_WM_RBUTTONDOWN,
	EV_WM_LBUTTONDBLCLK,
	EV_WM_KEYDOWN,
	EV_WM_KEYUP,
	EV_WM_SETFOCUS,
	EV_VN_MIRRORSUBTREE,
	EV_VN_GLOBALFOCUS,

	EV_COMMAND(CM_PREVNODE,       PrevNode),
	EV_COMMAND(CM_NEXTNODE,       NextNode),
	EV_COMMAND(CM_PREVHOLE,       PrevHole),
	EV_COMMAND(CM_NEXTHOLE,       NextHole),
	EV_COMMAND(CM_DELETE,         Delete),
	EV_COMMAND(CM_CUT,            Cut),
	EV_COMMAND(CM_COPY,           Copy),
	EV_COMMAND(CM_PASTE,          Paste),
	EV_COMMAND(CM_PARENT,         Parent),
	EV_COMMAND(CM_FOLD,           ToggleFold),
	EV_COMMAND(CM_ADDHOLETOLIST,  AddHoleToList),
	EV_COMMAND_AND_ID(CM_ADDHOLEBEFORE,  AddHoleBeforeAfter),
	EV_COMMAND_AND_ID(CM_ADDHOLEAFTER,   AddHoleBeforeAfter),
END_RESPONSE_TABLE;

const char far* TScrolledTreeWindowView<TFlowViewNode>::StaticName()
{
	return "Flow View";
}




DEFINE_DOC_TEMPLATE_CLASS(TStructDocument, TFlowView, TFlowTemplate);
TFlowTemplate flowTpl("Flow View", "*.STR", 0, "STR",
							dtAutoDelete | dtUpdateDir |dtAutoOpen);

const int ImageOffsetX = 8, ImageOffsetY = ArrowLength;

void TFlowView::Paint(TDC& dc, BOOL, TRect&)
{
	dc.SetBkMode(TRANSPARENT);
	dc.SelectObject(Pen);
	dc.SelectObject(FillBrush);
//	((TMyDC&)dc).Arrow(TPoint(ImageOffsetX+root->GetConnX(), ImageOffsetY), TMyDC::DOWN);
	TFlowViewNode::FlowTree = &Tree;
	if(Tree.GetRoot()) {
		Tree.GetRoot()->FlushNodePositions();
		Tree.GetRoot()->Draw((TMyDC&)dc, TPoint(ImageOffsetX, ImageOffsetY));
	}
}

