//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TListingView
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop

//#pragma option -Jgx

#include <owl\menu.h>

#include "views.rc"
#include "tree.h"
#include "strucdoc.h"
#include "mydc.h"
#include "viewnode.h"
#include "viewtree.h"
#include "scrwview.h"
#include "scrtrwvw.h"



class   TListingViewNode : public TViewNode {
public:
								 TListingViewNode(int n = 0, int fixed = FALSE);
	inline TListingViewNode *GetChild(int n) const;
	void                  CalcSizeFolded(TMyDC&);
	void                  DrawFolded(TMyDC&, TPoint& pos);
	void						 CalcSizeString(TMyDC&, const char const *);
	void						 DrawString(TMyDC&, TPoint&, const char const *);
	void						 CalcSizeHole(TMyDC&);
	void						 DrawHole(TMyDC&, TPoint&);
	void						 DoPosFocusFolded(TMyDC& dc, TPoint& pos);
	void  					 CalcSizeDelimiters(TMyDC& dc, \
														 char *start, char *between, char *end);

	void  					 DrawDelimiters(TMyDC& dc, TPoint& pos, \
														 char *start, char *between, char *end);
	static TViewTree<TListingViewNode>     *ListingTree;
};

inline TListingViewNode* TListingViewNode::GetChild(int n) const
{  return (TListingViewNode*)TTreeNode::GetChild(n);
}


const int         Indent  = 20;
const int         DistX   = 4;
#define  FocusColor   TColor(128, 128, 128)
static TBrush             FocusBrush  = TBrush(FocusColor);

TViewTree<TListingViewNode> *TListingViewNode::ListingTree = 0;

								 TListingViewNode::TListingViewNode(int n, int fixed):
	TViewNode(n, fixed)
{
}
void				    TListingViewNode::CalcSizeString(TMyDC& dc, const char const *s)
{
	if(GetFolded())
		CalcSizeFolded(dc);
	else
		SetNodeSize(dc.GetTextExtent(s, strlen(s)));
}

void				    		 TListingViewNode::DrawString(TMyDC& dc, TPoint& pos, const char const *s)
{
	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		dc.TextOut(pos, s);
	}
}
void						TListingViewNode::DoPosFocusFolded(TMyDC& dc, TPoint& pos)
{
	AddNodePosition(pos);
	if(ListingTree->GetFocus() == this)
		dc.FillRect(TRect(pos, NodeSize()), FocusBrush);
	if(GetFolded())
		DrawFolded(dc, pos);
	FlushChildrenNodePositions();
}
void  					TListingViewNode::CalcSizeFolded(TMyDC& dc)
{
	char	str[80];

	wsprintf(str, "%s <%s>", GetOpName(), GetSortName());
	SetNodeSize(dc.GetTextExtent(str, strlen(str))+TSize(PlusSize+DistX, 0));
}

void  					TListingViewNode::DrawFolded(TMyDC& dc, TPoint& pos)
{
	char	str[80];

	wsprintf(str, "%s <%s>", GetOpName(), GetSortName());
	dc.Plus(pos+TSize(0, (NodeSize().cy-PlusSize)/2));
	dc.TextOut(pos.x+PlusSize+DistX, pos.y, str);
}


void						TListingViewNode::CalcSizeHole(TMyDC& dc)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	CalcSizeString(dc, s);
}

void						TListingViewNode::DrawHole(TMyDC& dc, TPoint& pos)
{
	static char		s[40] = "? <";
	char	*p = GetSortName();
	strcpy(s+3, p); strcpy(s+3+strlen(p), ">");
	DrawString(dc, pos, s);
}

void  TPROGHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPROGHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TDECHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TDECSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TDECSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TNDECHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TNDECHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TIDHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TIDHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TTYPEHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TTYPEHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TSTATHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TSTATSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TSTATSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TVARSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TVARSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TEXPRHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TEXPRHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS1HOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS1HOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALTS2HOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALTS2HOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TALT2HOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TALT2HOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TLABELHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TLABELHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPBODYHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPBODYHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPPARHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPPARHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TPARGHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TPARGHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFBODYHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFBODYHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFPARHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFPARHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }
void  TFARGSHOLE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeHole(dc); }
void  TFARGSHOLE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawHole(dc, pos); }


void  TPROG<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TListingViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetStatements(); stats->CalcSize(dc);

		TSize    opensize, barsize, closesize;
		dc.GetTextExtent("|[", 2, opensize);
		dc.GetTextExtent("|",  1, barsize);
		dc.GetTextExtent("]|", 2, closesize);

		SetNodeSize(TSize(Indent+max(decs->NodeSize().cx, stats->NodeSize().cx),
								decs->NodeSize().cy+stats->NodeSize().cy+closesize.cy));
	}
}


void  TPROG<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;

		TSortDECS<TListingViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TListingViewNode>*  stats = GetStatements();
		dc.TextOut(pos, "|[");
		decs->Draw(dc, TPoint(pos.x+Indent, y));
		y += decs->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "]|");
	}
}


void  TDECS<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    semicolonsize;
		dc.GetTextExtent(";",     1, semicolonsize);

		int width       = 0;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortDEC<TListingViewNode> *dec   = GetDeclaration(i);
			dec->CalcSize(dc);
			int sw = dec->NodeSize().cx+semicolonsize.cx;
			if(sw>width) width = sw;
			height += dec->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}
void  TDECS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int y = pos.y;
		int numchildren = GetNumChildren();
		TSize    semicolonsize;
		dc.GetTextExtent(";",     1, semicolonsize);

		for(int i = 0; i<numchildren; i++) {
			TSortDEC<TListingViewNode> *dec   = GetDeclaration(i);
			if(i) dc.TextOut(pos.x, y, ";");
			dec->Draw(dc, TPoint(pos.x+semicolonsize.cx, y));
			y += dec->NodeSize().cy;
		}
	}
}

void  TListingViewNode::CalcSizeDelimiters(TMyDC& dc,
														 char *start, char *between, char *end)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int width       = startsize.cx;
		int height      = max(max(startsize.cy, betweensize.cy), endsize.cy);
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TListingViewNode *child   = GetChild(i);
			child->CalcSize(dc);
			width += child->NodeWidth() +(i<numchildren-1?betweensize.cx:0);
			if(child->NodeHeight() > height)
				height = child->NodeHeight();
		}
		width += endsize.cx;
		SetNodeSize(TSize(width, height));
	}
}
void  TListingViewNode::DrawDelimiters(TMyDC& dc, TPoint& pos,
														 char *start, char *between, char *end)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSize    startsize(0,0), betweensize(0,0), endsize(0,0);
		if(start)  dc.GetTextExtent(start, 		strlen(start), 	startsize);
		if(between)dc.GetTextExtent(between, 	strlen(between), 	betweensize);
		if(end)    dc.GetTextExtent(end, 		strlen(end), 		endsize);

		int numchildren = GetNumChildren();

		if(start)dc.TextOut(pos, start);
		int x = pos.x + startsize.cx;
		for(int i = 0; i<numchildren; i++) {
			TListingViewNode *child   = GetChild(i);
			child->Draw(dc, TPoint(x, pos.y));
			x += child->NodeWidth();
			if(i<numchildren-1 && between) {
				dc.TextOut(x, pos.y, between); x += betweensize.cx;
			}
		}
		if(end)dc.TextOut(x, pos.y, end);
	}
}


void  TNDECS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "var  ", ", ", 0);
}
void  TNDECS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "var  ", ", ", 0);
}
void  TNDEC<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc,  0, ":", 0);
}
void  TNDEC<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ":", 0);
}

void  TPROCDEC<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TListingViewNode>	 *id = GetProcName();
		id->CalcSize(dc);
		TSortPBODY<TListingViewNode> *pbody = GetBody();
		pbody->CalcSize(dc);

		TSize    procsize, eqsize;
		dc.GetTextExtent("proc ", 5, procsize);
		dc.GetTextExtent(" = ",   3, eqsize);

		SetNodeSize(TSize(procsize.cx + id->NodeWidth() + eqsize.cx + pbody->NodeWidth(),
								pbody->NodeHeight()));
	}
}

void  TPROCDEC<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSize    procsize, eqsize;
		dc.GetTextExtent("proc ", 5, procsize);
		dc.GetTextExtent(" = ",   3, eqsize);
		dc.TextOut(pos, "proc");
		int	x = pos.x+procsize.cx;
		TSortID<TListingViewNode>	 *id = GetProcName();
		id->Draw(dc, TPoint(x, pos.y));
		x += id->NodeWidth(); dc.TextOut(x, pos.y, " = ");
		x += eqsize.cx;
		TSortPBODY<TListingViewNode> *pbody = GetBody();
		pbody->Draw(dc, TPoint(x, pos.y));
	}
}


void  TFUNCDEC<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TListingViewNode>	 *id = GetFuncName();
		id->CalcSize(dc);
		TSortFBODY<TListingViewNode> *fbody = GetBody();
		fbody->CalcSize(dc);

		TSize    FUNCsize, eqsize;
		dc.GetTextExtent("func ", 5, FUNCsize);
		dc.GetTextExtent(" = ",   3, eqsize);

		SetNodeSize(TSize(FUNCsize.cx + id->NodeWidth() + eqsize.cx + fbody->NodeWidth(),
								fbody->NodeHeight()));
	}
}

void  TFUNCDEC<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSize    FUNCsize, eqsize;
		dc.GetTextExtent("func ", 5, FUNCsize);
		dc.GetTextExtent(" = ",   3, eqsize);
		dc.TextOut(pos, "func");
		int	x = pos.x+FUNCsize.cx;
		TSortID<TListingViewNode>	 *id = GetFuncName();
		id->Draw(dc, TPoint(x, pos.y));
		x += id->NodeWidth(); dc.TextOut(x, pos.y, " = ");
		x += eqsize.cx;
		TSortFBODY<TListingViewNode> *fbody = GetBody();
		fbody->Draw(dc, TPoint(x, pos.y));
	}
}

void  TID<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TID<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TINTTYPE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "int"); }
void  TINTTYPE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "int"); }
void  TCHARTYPE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "char"); }
void  TCHARTYPE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "char"); }
void  TREALTYPE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "real"); }
void  TREALTYPE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "real"); }
void  TBOOLTYPE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "bool"); }
void  TBOOLTYPE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "bool"); }

void  TSTATS<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    semicolonsize;
		dc.GetTextExtent(";", 1, semicolonsize);

		int width       = 0;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TListingViewNode> *stat   = GetStatement(i);
			stat->CalcSize(dc);
			int sw = semicolonsize.cx+stat->NodeSize().cx;
			if(sw>width)
				width = sw;
			height += stat->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}

void  TSTATS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSize    semicolonsize;
		dc.GetTextExtent(";", 1, semicolonsize);
		int y = pos.y;
		int numchildren = GetNumChildren();
		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TListingViewNode>  *stat = GetStatement(i);
			if(i)dc.TextOut(pos.x, y, ";");
			stat->Draw(dc, TPoint(pos.x+semicolonsize.cx, y));
			y += stat->NodeSize().cy;
		}
	}
}

void  TSKIP<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, "skip"); }
void  TSKIP<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, "skip"); }

void  TCONCASSIGN<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, " := ", 0);
}
void  TCONCASSIGN<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, " := ", 0);
}


void  TPROCCALL<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		GetProcName()->CalcSize(dc);
		SetNodeSize(GetProcName()->NodeSize());
	}
}
void  TPROCCALL<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded())
		GetProcName()->Draw(dc, pos);
}

void  TPROCCALLARGS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, "(", ")");
}
void  TPROCCALLARGS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, "(", ")");
}
void  TIFTHEN<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TListingViewNode>  *guard = GetGuard(); guard->CalcSize(dc);
		TSortSTATS<TListingViewNode> *stats = GetTrueStatements(); stats->CalcSize(dc);
		TSize    ifsize, thensize, fisize;
		dc.GetTextExtent("if ", 3, ifsize);
		dc.GetTextExtent(" then", 5, thensize);
		dc.GetTextExtent("fi ", 3, fisize);
		SetNodeSize(TSize(max((int)(ifsize.cx+guard->NodeWidth()+thensize.cx), Indent+stats->NodeWidth()),
								ifsize.cy+stats->NodeHeight()+fisize.cy));
	}
}

void  TIFTHEN<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSortEXPR<TListingViewNode>  *guard = GetGuard();
		TSortSTATS<TListingViewNode> *stats = GetTrueStatements();
		TSize    ifsize, thensize, fisize;
		dc.GetTextExtent("if ", 3, ifsize);
		dc.GetTextExtent(" then", 5, thensize);
		dc.GetTextExtent("fi ", 3, fisize);
		int x = pos.x+ifsize.cx;
		dc.TextOut(pos, "if ");
		guard->Draw(dc, TPoint(x, pos.y));
		x += guard->NodeWidth();
		dc.TextOut(x, pos.y, " then");
		int y = pos.y+ifsize.cy;
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "fi");
	}
}

void  TIFTHENELSE<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TListingViewNode>  *guard = GetGuard(); guard->CalcSize(dc);
		TSortSTATS<TListingViewNode> *tstats = GetTrueStatements(); tstats->CalcSize(dc);
		TSortSTATS<TListingViewNode> *fstats = GetFalseStatements(); fstats->CalcSize(dc);
		TSize    ifsize, thensize, elsesize, fisize;
		dc.GetTextExtent("if ", 3, ifsize);
		dc.GetTextExtent(" then", 5, thensize);
		dc.GetTextExtent("else", 4, elsesize);
		dc.GetTextExtent("fi", 2, fisize);
		SetNodeSize(TSize(max((int)(ifsize.cx+guard->NodeWidth()+thensize.cx),
									 Indent+max(tstats->NodeWidth(), fstats->NodeWidth())),
								ifsize.cy+tstats->NodeHeight()+elsesize.cy+fstats->NodeHeight()+fisize.cy));
	}
}

void  TIFTHENELSE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSortEXPR<TListingViewNode>  *guard = GetGuard();
		TSortSTATS<TListingViewNode> *tstats = GetTrueStatements();
		TSortSTATS<TListingViewNode> *fstats = GetFalseStatements();
		TSize    ifsize, thensize, elsesize, fisize;
		dc.GetTextExtent("if ", 3, ifsize);
		dc.GetTextExtent(" then", 5, thensize);
		dc.GetTextExtent("else", 4, elsesize);
		dc.GetTextExtent("fi", 2, fisize);
		dc.TextOut(pos, "if ");
		int x = pos.x+ifsize.cx;
		guard->Draw(dc, TPoint(x, pos.y));
		x += guard->NodeWidth();
		dc.TextOut(x, pos.y, " then");
		int y = pos.y+ifsize.cy;
		tstats->Draw(dc, TPoint(pos.x+Indent, y));
		y += tstats->NodeHeight();
		dc.TextOut(pos.x, y, "else");
		y += elsesize.cy;
		fstats->Draw(dc, TPoint(pos.x+Indent, y));
		y += fstats->NodeHeight();
		dc.TextOut(pos.x, y, "fi");
	}
}


void  TWHILE<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TListingViewNode>  *guard = GetGuard(); guard->CalcSize(dc);
		TSortSTATS<TListingViewNode> *stats = GetStatements(); stats->CalcSize(dc);
		TSize    whilesize, dosize, odsize;
		dc.GetTextExtent("while ", 6, whilesize);
		dc.GetTextExtent(" do", 3, dosize);
		dc.GetTextExtent("od", 2, odsize);
		SetNodeSize(TSize(max((int)(whilesize.cx+guard->NodeWidth()+dosize.cx), Indent+stats->NodeWidth()),
								whilesize.cy+stats->NodeHeight()+odsize.cy));
	}
}

void  TWHILE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSortEXPR<TListingViewNode>  *guard = GetGuard();
		TSortSTATS<TListingViewNode> *stats = GetStatements();
		TSize    whilesize, dosize, odsize;
		dc.GetTextExtent("while ", 6, whilesize);
		dc.GetTextExtent(" do", 3, dosize);
		dc.GetTextExtent("od", 2, odsize);
		int x = pos.x+whilesize.cx;
		dc.TextOut(pos, "while ");
		guard->Draw(dc, TPoint(x, pos.y));
		x += guard->NodeWidth();
		dc.TextOut(x, pos.y, " do");
		int y = pos.y+whilesize.cy;
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "od");
	}
}



void  TREAD<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "read(", ",", ")");
}
void  TREAD<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "read(", ",", ")");
}

void  TWRITE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "write(", ",", ")");
}
void  TWRITE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "write(", ",", ")");
}

void  TFOR<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortID<TListingViewNode>  	 *id  	= GetVariable(); 			id->CalcSize(dc);
		TSortEXPR<TListingViewNode>  *from  	= GetFromExpression(); 	from->CalcSize(dc);
		TSortEXPR<TListingViewNode>  *by  	= GetByExpression(); 	by->CalcSize(dc);
		TSortEXPR<TListingViewNode>  *to  	= GetToExpression(); 	to->CalcSize(dc);
		TSortSTATS<TListingViewNode> *stats = GetStatements(); 		stats->CalcSize(dc);
		TSize    forsize, fromsize, bysize, tosize, dosize, odsize;
		dc.GetTextExtent("for ",   4, forsize);
		dc.GetTextExtent(" from ", 6, fromsize);
		dc.GetTextExtent(" by ",   4, bysize);
		dc.GetTextExtent(" to ",   4, tosize);
		dc.GetTextExtent(" do", 	3, dosize);
		dc.GetTextExtent("od", 		2, odsize);
		SetNodeSize(TSize(max((int)(forsize.cx+id->NodeWidth()+
									 fromsize.cx+from->NodeWidth()+
									 bysize.cx + by->NodeWidth()+
									 tosize.cx + to->NodeWidth()+dosize.cx),
									 Indent+stats->NodeWidth()),
								forsize.cy+stats->NodeHeight()+odsize.cy));
	}
}

void  TFOR<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSortID<TListingViewNode>  	 *id  	= GetVariable();
		TSortEXPR<TListingViewNode>  *from  	= GetFromExpression();
		TSortEXPR<TListingViewNode>  *by  	= GetByExpression();
		TSortEXPR<TListingViewNode>  *to  	= GetToExpression();
		TSortSTATS<TListingViewNode> *stats = GetStatements();
		TSize    forsize, fromsize, bysize, tosize, dosize, odsize;
		dc.GetTextExtent("for ",   4, forsize);
		dc.GetTextExtent(" from ", 6, fromsize);
		dc.GetTextExtent(" by ",   4, bysize);
		dc.GetTextExtent(" to ",   4, tosize);
		dc.GetTextExtent(" do", 	3, dosize);
		dc.GetTextExtent("od", 		2, odsize);

		dc.TextOut(pos, "for ");
		int x = pos.x+forsize.cx;
		id->Draw(dc, TPoint(x, pos.y));
		x += id->NodeWidth();
		dc.TextOut(x, pos.y, " from ");
		x += fromsize.cx;
		from->Draw(dc, TPoint(x, pos.y));
		x += from->NodeWidth();
		dc.TextOut(x, pos.y, " by ");
		x += bysize.cx;
		by->Draw(dc, TPoint(x, pos.y));
		x += by->NodeWidth();
		dc.TextOut(x, pos.y, " to ");
		x += tosize.cx;
		to->Draw(dc, TPoint(x, pos.y));
		x += to->NodeWidth();
		dc.TextOut(x, pos.y, " do");

		int y = pos.y+forsize.cy;
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "od");
	}
}

void  TBLOCK<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TListingViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetStatements(); stats->CalcSize(dc);

		TSize    opensize, barsize, closesize;
		dc.GetTextExtent("|[", 2, opensize);
		dc.GetTextExtent("|",  1, barsize);
		dc.GetTextExtent("]|", 2, closesize);

		SetNodeSize(TSize(Indent+max(decs->NodeSize().cx, stats->NodeSize().cx),
								decs->NodeSize().cy+stats->NodeSize().cy+closesize.cy));
	}
}


void  TBLOCK<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;

		TSortDECS<TListingViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TListingViewNode>*  stats = GetStatements();
		dc.TextOut(pos, "|[");
		decs->Draw(dc, TPoint(pos.x+Indent, y));
		y += decs->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "]|");
	}
}


void  TVARS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TVARS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TEXPRS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc,  0, ", ", 0);
}
void  TEXPRS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TVARVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		GetVariable()->CalcSize(dc); SetNodeSize(GetVariable()->NodeSize());
	}
}

void  TVARVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded())
		GetVariable()->Draw(dc, pos);
}
void  TFUNCCALL<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, 0, 0);
}
void  TFUNCCALL<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, 0, 0);
}

void  TFUNCCALLARGS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, "(", ")");
}
void  TFUNCCALLARGS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, "(", ")");
}

void  TNUMVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TNUMVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TCHARVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TCHARVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TIMPLICATION<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " => ", ")");
}
void  TIMPLICATION<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " => ", ")");
}
void  TEQUIVALENCE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " == ", ")");
}
void  TEQUIVALENCE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " == ", ")");
}

void  TOR<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " \\/ ", ")");
}
void  TOR<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " \\/ ", ")");
}

void  TAND<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " /\\ ", ")");
}
void  TAND<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " /\\ ", ")");
}

void  TLESS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " < ", ")");
}
void  TLESS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " < ", ")");
}


void  TLESSEQ<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " <= ", ")");
}
void  TLESSEQ<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " <= ", ")");
}


void  TGREATER<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " > ", ")");
}
void  TGREATER<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " > ", ")");
}


void  TGREATEREQ<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " >= ", ")");
}
void  TGREATEREQ<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " >= ", ")");
}


void  TEQUALITY<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " == ", ")");
}
void  TEQUALITY<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " == ", ")");
}

void  TDIFFERENCE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " /= ", ")");
}
void  TDIFFERENCE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " /= ", ")");
}

void  TADDITION<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " + ", ")");
}
void  TADDITION<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " + ", ")");
}

void  TSUBTRACTION<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " - ", ")");
}
void  TSUBTRACTION<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " - ", ")");
}

void  TMULTIPLICATION<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", "*", ")");
}
void  TMULTIPLICATION<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", "*", ")");
}

void  TDIVISION<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", "/", ")");
}
void  TDIVISION<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", "/", ")");
}

void  TMODULO<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "(", " mod ", ")");
}
void  TMODULO<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "(", " mod ", ")");
}


void  TUNARYMINUS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "-", 0, 0);
}
void  TUNARYMINUS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "-", 0, 0);
}

void  TNOT<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "NOT ", 0, 0);
}
void  TNOT<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "NOT ", 0, 0);
}

void  TEXPRBLOCK<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortDECS<TListingViewNode>*   decs  = GetDeclarations(); decs->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetStatements();   stats->CalcSize(dc);
		TSortEXPR<TListingViewNode>*   expr  = GetExpression();   expr->CalcSize(dc);

		TSize    opensize, barsize, closesize;
		dc.GetTextExtent("|[", 2, opensize);
		dc.GetTextExtent("|",  1, barsize);
		dc.GetTextExtent("]|", 2, closesize);

		SetNodeSize(TSize(Indent+max((int)(max(decs->NodeSize().cx, stats->NodeSize().cx)), expr->NodeWidth()),
								decs->NodeSize().cy+stats->NodeSize().cy+expr->NodeHeight()+closesize.cy));
	}
}
void  TEXPRBLOCK<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;

		TSortDECS<TListingViewNode>*   decs  = GetDeclarations();
		TSortSTATS<TListingViewNode>*  stats = GetStatements();
		TSortEXPR<TListingViewNode>*   expr  = GetExpression();

		dc.TextOut(pos, "|[");
		decs->Draw(dc, TPoint(pos.x+Indent, y));
		y += decs->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		expr->Draw(dc, TPoint(pos.x+Indent, y));
		y += expr->NodeHeight();
		dc.TextOut(pos.x, y, "]|");
	}
}

void  TPBODYSTATS<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSize    semicolonsize;
		dc.GetTextExtent(";", 1, semicolonsize);

		int width       = 0;
		int height      = semicolonsize.cy; // for closing bracket
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TListingViewNode> *stat   = GetStatement(i);
			stat->CalcSize(dc);
			int sw = Indent+semicolonsize.cx+stat->NodeSize().cx;
			if(sw>width)
				width = sw;
			height += stat->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}

void  TPBODYSTATS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		TSize    semicolonsize;
		dc.GetTextExtent(";", 1, semicolonsize);
		TEXTMETRIC tm;
		dc.GetTextMetrics(tm);
		int y = pos.y;
		int numchildren = GetNumChildren();
		dc.TextOut(pos, "(");
		for(int i = 0; i<numchildren; i++) {
			TSortSTAT<TListingViewNode>  *stat = GetStatement(i);
			if(i)
				dc.TextOut(pos.x+Indent, y, ";");
			stat->Draw(dc, TPoint(pos.x+Indent+semicolonsize.cx, y));
			y += stat->NodeSize().cy;
		}
		dc.TextOut(pos.x, y, ")");
	}
}


void  TPBODYPARSSTATS<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortPPARS<TListingViewNode>*  pars  = GetProcParameters(); pars->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetStatements(); stats->CalcSize(dc);

		TSize    opensize, barsize, closesize;
		dc.GetTextExtent("(", 1, opensize);
		dc.GetTextExtent("|", 1, barsize);
		dc.GetTextExtent(")", 1, closesize);

		SetNodeSize(TSize(Indent+max(pars->NodeSize().cx, stats->NodeSize().cx),
								pars->NodeSize().cy+stats->NodeSize().cy+closesize.cy));
	}
}
void  TPBODYPARSSTATS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;

		TSortPPARS<TListingViewNode>*  pars  = GetProcParameters();
		TSortSTATS<TListingViewNode>*  stats = GetStatements();
		dc.TextOut(pos, "(");
		pars->Draw(dc, TPoint(pos.x+Indent, y));
		y += pars->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		stats->Draw(dc, TPoint(pos.x+Indent, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, ")");
	}
}

void  TPPARS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TPPARS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TPPARVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TPPARVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TPPARREF<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "! ", ": ", 0);
}
void  TPPARREF<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "! ", ": ", 0);
}

void  TPARGS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TPARGS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TPARGVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TPARGVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TPARGREF<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "! ", ": ", 0);
}
void  TPARGREF<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "! ", ": ", 0);
}





void  TFBODY<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortFPARS<TListingViewNode>*  pars  = GetFuncParameters(); pars->CalcSize(dc);
		TSortEXPR<TListingViewNode>*   expr  = GetExpression(); 		expr->CalcSize(dc);

		TSize    opensize, barsize, closesize;
		dc.GetTextExtent("(", 1, opensize);
		dc.GetTextExtent("|", 1, barsize);
		dc.GetTextExtent(")", 1, closesize);

		SetNodeSize(TSize(Indent+max(pars->NodeSize().cx, expr->NodeSize().cx),
								pars->NodeSize().cy+expr->NodeSize().cy+closesize.cy));
	}
}
void  TFBODY<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;

		TSortFPARS<TListingViewNode>*  pars  = GetFuncParameters();
		TSortEXPR<TListingViewNode>*   expr  = GetExpression();

		dc.TextOut(pos, "(");
		pars->Draw(dc, TPoint(pos.x+Indent, y));
		y += pars->NodeHeight();
		dc.TextOut(pos.x, y, "|");
		expr->Draw(dc, TPoint(pos.x+Indent, y));
		y += expr->NodeHeight();
		dc.TextOut(pos.x, y, ")");
	}
}



void  TFPARS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ", ", 0);
}
void  TFPARS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ", ", 0);
}

void  TFPARVALUE<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ": ", 0);
}
void  TFPARVALUE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ": ", 0);
}

void  TFARGS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, "? ", ", ? ", 0);
}
void  TFARGS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, "? ", ", ? ", 0);
}

void	TNUMCASE<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TListingViewNode>*   expr  = GetCase();   expr->CalcSize(dc);
		TSortALTS1<TListingViewNode>*  alts  = GetIn(); alts->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetOut();   stats->CalcSize(dc);

		TSize    numcasesize, insize, esacsize;
		dc.GetTextExtent("numcase ", 	8, numcasesize);
		dc.GetTextExtent(" in",  		3, insize);
		dc.GetTextExtent("esac",  		4, esacsize);

		SetNodeSize(TSize(max(numcasesize.cx+expr->NodeWidth()+insize.cx,
									 Indent+max(alts->NodeWidth(), Indent+stats->NodeWidth())),
								expr->NodeHeight()+alts->NodeHeight()+stats->NodeHeight()+
								esacsize.cy));
	}
}

void  TNUMCASE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;
		TSortEXPR<TListingViewNode>*   expr  = GetCase();
		TSortALTS1<TListingViewNode>*  alts  = GetIn();
		TSortSTATS<TListingViewNode>*  stats = GetOut();
		TSize    numcasesize;
		dc.GetTextExtent("numcase ", 	8, numcasesize);

		dc.TextOut(pos, "numcase");
		expr->Draw(dc, TPoint(pos.x+numcasesize.cx, y));
		dc.TextOut(pos.x+numcasesize.cx+expr->NodeWidth(), y, " in");
		y += expr->NodeHeight();
		alts->Draw(dc, TPoint(pos.x+Indent, y));
		y += alts->NodeHeight();
		dc.TextOut(pos.x+Indent, y, "out");
		stats->Draw(dc, TPoint(pos.x+Indent*2, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "esac");
	}
}

void	TLABCASE<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		TSortEXPR<TListingViewNode>*   expr  = GetCase();   expr->CalcSize(dc);
		TSortALTS2<TListingViewNode>*  alts  = GetIn(); alts->CalcSize(dc);
		TSortSTATS<TListingViewNode>*  stats = GetOut();   stats->CalcSize(dc);

		TSize    labcasesize, insize, esacsize;
		dc.GetTextExtent("labcase ", 	8, labcasesize);
		dc.GetTextExtent(" in",  		3, insize);
		dc.GetTextExtent("esac",  		4, esacsize);

		SetNodeSize(TSize(max(labcasesize.cx+expr->NodeWidth()+insize.cx,
									 Indent+max(alts->NodeWidth(), Indent+stats->NodeWidth())),
								expr->NodeHeight()+alts->NodeHeight()+stats->NodeHeight()+
								esacsize.cy));
	}
}

void  TLABCASE<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int			y = pos.y;
		TSortEXPR<TListingViewNode>*   expr  = GetCase();
		TSortALTS2<TListingViewNode>*  alts  = GetIn();
		TSortSTATS<TListingViewNode>*  stats = GetOut();
		TSize    	labcasesize;
		dc.GetTextExtent("labcase ", 	8, labcasesize);

		dc.TextOut(pos, "labcase");
		expr->Draw(dc, TPoint(pos.x+labcasesize.cx, y));
		dc.TextOut(pos.x+labcasesize.cx+expr->NodeWidth(), y, " in");
		y += expr->NodeHeight();
		alts->Draw(dc, TPoint(pos.x+Indent, y));
		y += alts->NodeHeight();
		dc.TextOut(pos.x+Indent, y, "out");
		stats->Draw(dc, TPoint(pos.x+Indent*2, y));
		y += stats->NodeHeight();
		dc.TextOut(pos.x, y, "esac");
	}
}

void  TALTS1<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		int width       = Indent;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortSTATS<TListingViewNode> *stats   = GetStatements(i);
			stats->CalcSize(dc);
			int sw = Indent+stats->NodeSize().cx;
			if(sw>width)
				width = sw;
			height += stats->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}

void  TALTS1<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int y = pos.y;
		int numchildren = GetNumChildren();
		for(int i = 0; i<numchildren; i++) {
			TSortSTATS<TListingViewNode>  *stats = GetStatements(i);
			if(i)dc.TextOut(pos.x, y, "[]");
			stats->Draw(dc, TPoint(Indent+pos.x, y));
			y += stats->NodeSize().cy;
		}
	}
}

void  TALTS2<TListingViewNode>::CalcSize(TMyDC& dc)
{
	if(GetFolded()) {
		CalcSizeFolded(dc);
	} else {
		int width       = Indent;
		int height      = 0;
		int numchildren = GetNumChildren();

		for(int i = 0; i<numchildren; i++) {
			TSortALT2<TListingViewNode> *alt2   = GetAlternative(i);
			alt2->CalcSize(dc);
			int sw = Indent+alt2->NodeSize().cx;
			if(sw>width)
				width = sw;
			height += alt2->NodeSize().cy;
		}
		SetNodeSize(TSize(width, height));
	}
}

void  TALTS2<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DoPosFocusFolded(dc, pos);
	if(!GetFolded()) {
		int y = pos.y;
		int numchildren = GetNumChildren();
		for(int i = 0; i<numchildren; i++) {
			TSortALT2<TListingViewNode>  *alt2 = GetAlternative(i);
			if(i)dc.TextOut(pos.x, y, "[]");
			alt2->Draw(dc, TPoint(Indent+pos.x, y));
			y += alt2->NodeSize().cy;
		}
	}
}
void  TALT2<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ": ", 0);
}
void  TALT2<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ": ", 0);
}
void  TLABELS<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeDelimiters(dc, 0, ",", 0);
}
void  TLABELS<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawDelimiters(dc, pos, 0, ",", 0);
}

void  TNUMLABEL<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TNUMLABEL<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }

void  TCHARLABEL<TListingViewNode>::CalcSize(TMyDC& dc)
{	CalcSizeString(dc, (char *)GetData()); }
void  TCHARLABEL<TListingViewNode>::Draw(TMyDC& dc, TPoint& pos)
{	DrawString(dc, pos, (char *)GetData()); }



typedef TScrolledTreeWindowView<TListingViewNode> TListingView;

DEFINE_RESPONSE_TABLE1(TListingView, TScrolledWindowView)
	EV_WM_CHAR,
	EV_WM_LBUTTONDOWN,
	EV_WM_RBUTTONDOWN,
	EV_WM_LBUTTONDBLCLK,
   EV_WM_KEYDOWN,
   EV_WM_KEYUP,
	EV_WM_SETFOCUS,
	EV_VN_MIRRORSUBTREE,
   EV_VN_GLOBALFOCUS,

   EV_COMMAND(CM_PREVNODE,       PrevNode),
   EV_COMMAND(CM_NEXTNODE,       NextNode),
   EV_COMMAND(CM_PREVHOLE,       PrevHole),
   EV_COMMAND(CM_NEXTHOLE,       NextHole),
   EV_COMMAND(CM_DELETE,         Delete),
   EV_COMMAND(CM_CUT,            Cut),
	EV_COMMAND(CM_COPY,           Copy),
	EV_COMMAND(CM_PASTE,          Paste),
	EV_COMMAND(CM_PARENT,         Parent),
	EV_COMMAND(CM_FOLD,           ToggleFold),
	EV_COMMAND(CM_ADDHOLETOLIST,  AddHoleToList),
	EV_COMMAND_AND_ID(CM_ADDHOLEBEFORE,  AddHoleBeforeAfter),
	EV_COMMAND_AND_ID(CM_ADDHOLEAFTER,   AddHoleBeforeAfter),
END_RESPONSE_TABLE;

const char far* TScrolledTreeWindowView<TListingViewNode>::StaticName()
{
	return "Listing View";
}

void TListingView::Paint(TDC& dc, BOOL, TRect&)
{
   dc.SetBkMode(TRANSPARENT);
	TListingViewNode::ListingTree = &Tree;
	if(Tree.GetRoot()) {
		Tree.GetRoot()->FlushNodePositions();
		Tree.GetRoot()->Draw((TMyDC&)dc, TPoint(0, 0));
	}
}



DEFINE_DOC_TEMPLATE_CLASS(TStructDocument, TListingView, TListingTemplate);
TListingTemplate listingTpl("Listing View", "*.STR", 0, "STR",
									 dtAutoDelete | dtUpdateDir |dtAutoOpen);

