//----------------------------------------------------------------------------
//   Copyright 1995 Hugo Lyppens
//   Implements class TScrolledTreeWindowView
//----------------------------------------------------------------------------



template<class T>
class _DOCVIEWCLASS TScrolledTreeWindowView : public TScrolledWindowView
{
public:
	 TScrolledTreeWindowView(TStructDocument& doc, TWindow *parent = 0);
	 static const char far* StaticName();

	 //
	 // overridden virtuals from TView
	 LPCSTR   GetViewName(){ return StaticName();}

	 TSize    CalcSize();

protected:
	 virtual  char    GetOpKey( Operator op) { return 0; }
	 virtual  char   *GetOpMenuString( Operator op) { return ::GetOpName(op); }

	 void      Paint(TDC&, BOOL erase, TRect&);
	 LRESULT  EvCommand(UINT id, HWND hWndCtl, UINT notifyCode);

	 // Message response functions
	 BOOL      VnMirrorSubtree(TNodeAddress *address);
	 BOOL      VnGlobalFocus(TNodeAddress *address);
	 BOOL      VnNewActive(TNodeAddress *address);

	 void		  EvSetFocus(HWND);
	 void      EvLButtonDown(UINT ModKeys, TPoint& point);
	 void      EvRButtonDown(UINT ModKeys, TPoint& point);
	 void      EvLButtonDblClk(UINT ModKeys, TPoint& point);
	 void      EvChar(UINT key, UINT repeatcount, UINT flags);
	 void      EvKeyDown(UINT key, UINT repeatcount, UINT flags);
	 void      EvKeyUp(UINT key, UINT repeatcount, UINT flags);

	 void      ShowMenu(const TPoint&);
	 void      HandleButtonDown(TPoint& point, BOOL nofold = FALSE);
	 void      HandleLButtonDown(TPoint& point);
	 void      ToggleFold();

	 void      ChangeFocus(T *);
	 void      Parent();
	 void      PrevNode();
	 void      NextNode();
	 void      PrevHole();
	 void      NextHole();

	 void      AddHoleToList();
	 void      AddHoleBeforeAfter(WPARAM);
	 void      Delete();
	 void      Cut();
	 void      Copy();
	 void      Paste();

	 DECLARE_RESPONSE_TABLE(TScrolledTreeWindowView);

	 TViewTree<T>     Tree;
	 TStructDocument *StructDoc;  // same as Doc member, but cast to derived class
	 BOOL             ShiftDown;
};

NOTIFY_SIG(vnMirrorSubtree, TNodeAddress*);
NOTIFY_SIG(vnGlobalFocus,   TNodeAddress*);
NOTIFY_SIG(vnNewActive,     TNodeAddress*);

#define EV_VN_MIRRORSUBTREE VN_DEFINE(vnMirrorSubtree, VnMirrorSubtree, pointer)
#define EV_VN_GLOBALFOCUS   VN_DEFINE(vnGlobalFocus,   VnGlobalFocus,   pointer)
#define EV_VN_NEWACTIVE     VN_DEFINE(vnNewActive,     VnNewActive,     pointer)

template<class T>
void     TScrolledTreeWindowView<T>::EvSetFocus(HWND)
{
	static TNodeAddress	address;

	if(Tree.GetFocus()) {
		Tree.GetNodeAddress(Tree.GetFocus(), address);
		StructDoc->SetActiveAddress(address);
	}
}

template<class T>
void     TScrolledTreeWindowView<T>::EvKeyDown(UINT key, UINT , UINT )
{
	if(key ==  VK_SHIFT)
		ShiftDown = TRUE;
}


template<class T>
void     TScrolledTreeWindowView<T>::EvKeyUp(UINT key, UINT , UINT )
{
	if(key ==  VK_SHIFT)
		ShiftDown = FALSE;
}



template<class T>
void     TScrolledTreeWindowView<T>::Delete()
{
	static TNodeAddress address;
	T                  *node = Tree.GetFocus();

	if(node->GetOperator() >= OP_FIRST) {
		Tree.GetNodeAddress(node, address);
		StructDoc->Delete(address);
		Tree.SetFocus(Tree.NodeAt(address));
   } else {
      int cn = Tree.GetMyChildNumber(node);
      node = Tree.GetParent(node);
      if(node && node->IsList()) {
			Tree.GetNodeAddress(node, address);
         StructDoc->ListDelete(address, cn);
         node = Tree.NodeAt(address);
         int numchildren = Tree.GetNumChildren(node);
         if(numchildren>0) {
            if(cn >= numchildren)
               cn  = numchildren-1;
            node = Tree.GetChild(node, cn);
         }
         Tree.SetFocus(node);
			if(node->GetOperator() >= OP_FIRST) {
            NextHole();
			}
		}
	}
}

template<class T>
BOOL     TScrolledTreeWindowView<T>::VnMirrorSubtree(TNodeAddress *address)
{
	TTerm         *termnode = StructDoc->GetTree()->NodeAt(*address);
	T *viewnode = Tree.NodeAt(*address);
	T *curfocus = Tree.GetFocus();
	BOOL setfocus  = (curfocus->IsDescendant(viewnode)) ||
						  (curfocus == viewnode);

	Tree.CreateMirror(Tree.GetParent(viewnode),
							Tree.GetMyChildNumber(viewnode),
							termnode);
	AdjustScroller();
	if(setfocus)
		Tree.SetFocus(Tree.NodeAt(*address));
	Invalidate();
	return TRUE;
}

template<class T>
BOOL     TScrolledTreeWindowView<T>::VnGlobalFocus(TNodeAddress *address)
{
	ChangeFocus(Tree.NodeAt(*address));
	return TRUE;
}

template<class T>
BOOL     TScrolledTreeWindowView<T>::VnNewActive(TNodeAddress *)
{
	return TRUE;
}


template<class T>
void      TScrolledTreeWindowView<T>::Cut()
{
	 static TNodeAddress address;

	 Tree.GetNodeAddress(Tree.GetFocus(), address);
	 StructDoc->Cut(address, *this);
	 Tree.SetFocus(Tree.NodeAt(address));
}

template<class T>
void      TScrolledTreeWindowView<T>::Copy()
{
	 static TNodeAddress address;

    Tree.GetNodeAddress(Tree.GetFocus(), address);
    StructDoc->Copy(address, *this);
}

template<class T>
void      TScrolledTreeWindowView<T>::Paste()
{
    static TNodeAddress address;

    Tree.GetNodeAddress(Tree.GetFocus(), address);
	 StructDoc->Paste(address, *this);
    Tree.SetFocus(Tree.NodeAt(address));
}

template<class T>
void      TScrolledTreeWindowView<T>::Parent()
{
   T  *focus    = Tree.GetFocus();
   if(focus = Tree.GetParent(focus))
      ChangeFocus(focus);
}

template<class T>
void      TScrolledTreeWindowView<T>::NextHole()
{
	T  *focus    = Tree.GetFocus();
   T  *nexthole = focus;
   do {
      nexthole = Tree.GetNextNode(nexthole);
		if(!nexthole)
         nexthole = Tree.GetRoot();
   } while (nexthole != focus && (nexthole->GetOperator() >= OP_FIRST || !nexthole->IsVisible()));
   if(focus != nexthole)
      ChangeFocus(nexthole);
}

template<class T>
void      TScrolledTreeWindowView<T>::PrevHole()
{
   T  *focus    = Tree.GetFocus();
   T  *prevhole = focus;
   do {
      prevhole = Tree.GetPrevNode(prevhole);
      if(!prevhole)
         prevhole = Tree.GetLastNode();
	} while (prevhole != focus && (prevhole->GetOperator() >= OP_FIRST || !prevhole->IsVisible()));
   if(focus != prevhole)
      ChangeFocus(prevhole);
}

template<class T>
void      TScrolledTreeWindowView<T>::NextNode()
{
   T  *focus    = Tree.GetFocus();
	T  *nextnode = focus;

   do {
      nextnode = Tree.GetNextNode(nextnode);
      if(!nextnode)
         nextnode = Tree.GetRoot();
	} while(!nextnode->IsVisible());
   if(focus != nextnode)
      ChangeFocus(nextnode);
}

template<class T>
void      TScrolledTreeWindowView<T>::PrevNode()
{
   T  *focus    = Tree.GetFocus();
	T  *prevnode = focus;

   do {
      prevnode = Tree.GetPrevNode(prevnode);
		if(!prevnode)
			prevnode = Tree.GetLastNode();
   } while(!prevnode->IsVisible());
   if(focus != prevnode)
      ChangeFocus(prevnode);
}

template<class T>
void      TScrolledTreeWindowView<T>::EvChar(UINT key, UINT repeatcount, UINT flags)
{
   T  *focus = Tree.GetFocus();

	TWindow::EvChar(key, repeatcount, flags);
   switch(key) {
   case 13:
      ShowMenu(focus->GetNodePosition(0)+
		  TPoint(focus->NodeWidth()/2, focus->NodeHeight()/2)-
        TPoint(Scroller->XPos, Scroller->YPos));
      break;
   default:
       Sort     sort = focus->GetSort();
       Operator op   = focus->GetOperator();
		if(sort == op) {
         for(op = OP_FIRST; op <= OP_LAST; op++) {
            if(GetOpSort(op)==sort && key == GetOpKey(op)) {
					EvCommand(CM_REFINE+(int)op, 0, 0);
				}
			}
		}
   }
}


template<class T>
LRESULT  TScrolledTreeWindowView<T>::EvCommand(UINT id, HWND hWndCtl, UINT notifyCode)
{
      static     TNodeAddress address;
		T *newfocus;

      if(id>=CM_REFINE && id<=CM_REFINE+(int)OP_LAST) {
			Tree.GetNodeAddress(Tree.GetFocus(), address);
         StructDoc->Refine(address, ( Operator)(id-CM_REFINE), this);
			newfocus = Tree.NodeAt(address);
			if(newfocus->GetNumChildren()) {
            newfocus = Tree.GetChild(newfocus, 0);
         }
         Tree.SetFocus(newfocus);
		}
      return TWindow::EvCommand(id, hWndCtl, notifyCode);
}

template<class T>
void TScrolledTreeWindowView<T>::AddHoleToList()
{
   static   TNodeAddress     address;
   T     *node = Tree.GetFocus();
	Tree.GetNodeAddress(node, address);
   StructDoc->ListInsert(address, node->GetNumChildren());
   Tree.SetFocus(Tree.NodeAt(address));
}

template<class T>
void TScrolledTreeWindowView<T>::AddHoleBeforeAfter(WPARAM id)
{
   static   TNodeAddress     address;
   T     *node = Tree.GetFocus();
   int    cn   = Tree.GetMyChildNumber(node);
	while((node = Tree.GetParent(node)) && !node->IsList()) {
      cn = Tree.GetMyChildNumber(node);
   }
   if(node) {
      if(id == CM_ADDHOLEAFTER)
			cn++;
      Tree.GetNodeAddress(node, address);
      StructDoc->ListInsert(address, cn);
		node = Tree.NodeAt(address);
		Tree.SetFocus(Tree.GetChild(node, cn));
   }
}

template<class T>
TScrolledTreeWindowView<T>::TScrolledTreeWindowView(TStructDocument& doc, TWindow *parent):
					TScrolledWindowView(doc, parent), StructDoc(&doc), Tree(doc.GetTree()),
					ShiftDown(FALSE)
{
	static TNodeAddress	address;

	StructDoc->GetActiveAddress(address);
	Tree.SetFocus(Tree.NodeAt(address));
	Attr.AccelTable = IDA_FLOWVIEW;
	SetViewMenu(new TMenuDescr(IDM_FLOWVIEW,0,1,0,0,0,1));
	SetBkgndColor(TColor::LtGray);
}

template<class T>
void TScrolledTreeWindowView<T>::EvRButtonDown(UINT ModKeys, TPoint& point)
{
	TWindow::EvRButtonDown(ModKeys, point);


	TPoint          pt(Scroller->XPos, Scroller->YPos);
	T  *newfocus = Tree.WhichNode(point + pt);

	if(newfocus) {
		if(newfocus != Tree.GetFocus())
			ChangeFocus(newfocus);
	}
	ShowMenu(point);
}

template<class T>
void TScrolledTreeWindowView<T>::ShowMenu(const TPoint& point)
{
		TPoint         p    = point;
		TPopupMenu     m;
		T *node = Tree.GetFocus();

		 Sort     sort = node->GetSort();
		 Operator op   = node->GetOperator();
      char  s[50], *opname;

      if(sort == op) {
         for(op = OP_FIRST; op<=OP_LAST;
             op++) {
            if(GetOpSort(op)==sort) {
					m.AppendMenu(MF_ENABLED, CM_REFINE+(int)op, GetOpMenuString(op));
				}
			}
			m.AppendMenu(MF_SEPARATOR, 0, 0);
			m.AppendMenu(StructDoc->ClipboardOK(sort, *this)?
									MF_ENABLED:MF_GRAYED,
							 CM_PASTE,  "&Paste\tCtrl+V");

         T  *par = Tree.GetParent(node);
         if(par && par->IsList()) {
            m.AppendMenu(MF_SEPARATOR, 0, 0);
            wsprintf(s, "Delete from %s\tDelete", GetOpName(par->GetOperator()));
				m.AppendMenu(MF_ENABLED, CM_DELETE, s);
         }
      } else {
         m.AppendMenu(MF_ENABLED, CM_FOLD, node->GetFolded()?"Unfold\tF":"Fold\tF");
         m.AppendMenu(MF_SEPARATOR, 0, 0);
         m.AppendMenu(MF_ENABLED, CM_CUT,    "Cu&t\tCtrl+X");
			m.AppendMenu(MF_ENABLED, CM_COPY,   "&Copy\tCtrl+C");
         m.AppendMenu(MF_ENABLED, CM_DELETE, "&Delete\tDelete");
      }
      if(node->IsList()) {
			m.AppendMenu(MF_SEPARATOR, 0, 0);
         opname = GetOpName(node->GetOperator());
         wsprintf(s, "Add Hole to %s\tH", opname);
         m.AppendMenu(MF_ENABLED, CM_ADDHOLETOLIST,  s);
		}

      T  *par = node;
      while(par = Tree.GetParent(par)) {
			if(par->IsList()) {
            m.AppendMenu(MF_SEPARATOR, 0, 0);
            opname = GetOpName(par->GetOperator());
            wsprintf(s, "Add to ancestor %s Before\tB", opname);
            m.AppendMenu(MF_ENABLED, CM_ADDHOLEBEFORE, s);
            wsprintf(s, "Add to ancestor %s After\tA", opname);
            m.AppendMenu(MF_ENABLED, CM_ADDHOLEAFTER,  s);
            break;
         }
      }
      m.AppendMenu(MF_SEPARATOR, 0, 0);
		m.AppendMenu(MF_ENABLED, CM_NEXTHOLE,  "Next Hole\tDown");
		m.AppendMenu(MF_ENABLED, CM_PREVHOLE,  "Prev Hole\tUp");
		m.AppendMenu(MF_ENABLED, CM_NEXTNODE,  "Next Node\tRight");
		m.AppendMenu(MF_ENABLED, CM_PREVNODE,  "Prev Node\tLeft");
      m.AppendMenu(MF_ENABLED, CM_PARENT,    "Parent\tP");

      ClientToScreen(p);
      m.TrackPopupMenu(TPM_RIGHTBUTTON|TPM_LEFTALIGN, p,0,(HWND)*this);
}

template<class T>
void TScrolledTreeWindowView<T>::EvLButtonDblClk(UINT ModKeys, TPoint& point)
{
   TWindow::EvLButtonDblClk(ModKeys, point);
   HandleLButtonDown(point);
}

template<class T>
void TScrolledTreeWindowView<T>::EvLButtonDown(UINT ModKeys, TPoint& point)
{
	TWindow::EvLButtonDown(ModKeys, point);
   HandleLButtonDown(point);
}

template<class T>
void TScrolledTreeWindowView<T>::HandleLButtonDown(TPoint& point)
{
	static   TNodeAddress     address;

	HandleButtonDown(point, ShiftDown);
	if(ShiftDown) {
		Tree.GetNodeAddress(Tree.GetFocus(), address);
		StructDoc->GlobalFocus(address);
	}
}

template<class T>
void TScrolledTreeWindowView<T>::ChangeFocus(T *newfocus)
{
   T  *oldfocus = Tree.GetFocus();
	static   TNodeAddress     address;

	while(newfocus && !newfocus->IsVisible())
		newfocus = Tree.GetParent(newfocus);

	if(newfocus && newfocus!=oldfocus) {
		TPoint          pt(Scroller->XPos, Scroller->YPos);
		Tree.SetFocus(newfocus);
		int np, i;
		np = oldfocus->GetNumNodePositions();
		for(i = 0; i<np; i++) {
			InvalidateRect(TRect(oldfocus->GetNodePosition(i)-pt,
										oldfocus->NodeSize()));
		}
		np = newfocus->GetNumNodePositions();
		for(i = 0; i<np; i++) {
			InvalidateRect(TRect(newfocus->GetNodePosition(i)-pt,
										newfocus->NodeSize()));
		}
		Tree.GetNodeAddress(newfocus, address);
		StructDoc->SetActiveAddress(address);
	}
}

template<class T>
void TScrolledTreeWindowView<T>::ToggleFold()
{
   T  *focus = Tree.GetFocus();

   if(Tree.GetNumChildren(focus)>0) {
      focus->SetFolded(!focus->GetFolded());
      AdjustScroller();
      Invalidate();
   }
}

template<class T>
void TScrolledTreeWindowView<T>::HandleButtonDown(TPoint& point, BOOL nofold)
{
      TPoint          pt(Scroller->XPos, Scroller->YPos);
      T  *newfocus = Tree.WhichNode(point + pt);
      T  *oldfocus = Tree.GetFocus();

      if(newfocus) {
         if(newfocus != oldfocus) {
            ChangeFocus(newfocus);
         } else if(!nofold) {
            ToggleFold();
         }
      }
}

template<class T>
TSize    TScrolledTreeWindowView<T>::CalcSize()
{
   T   *root = Tree.GetRoot();

   if(root) {
      root->CalcSize((TMyDC&)TClientDC(HWindow));
		return root->NodeSize();
   }
   return TSize(0, 0);
}

