//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements PL4 signature
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop

#include "tree.h"
#include "strucdoc.h"


void		AskData(TWindow *window, TTerm *node)
{
	char		s[MAXOPDATALEN+1], v[20];
	char		h[MAXOPNAMELEN+51];

	s[0] = 0;
	wsprintf(h, "Operator %s additional info", node->GetOpName());
	switch(node->GetOperator()) {
	case OP_NUMVALUE:
	case OP_NUMLABEL:
		TInputDialog(window, h, "Enter numeric value:",
						 s, sizeof s).Execute();
		ostrstream(v, sizeof v)<<atof(s)<<'\0';
		node->SetData(v, strlen(v)+1);
		break;
	case OP_CHARVALUE:
	case OP_CHARLABEL:
		TInputDialog(window, h, "Enter a character:",
						 s, sizeof s).Execute();
		s[1] = s[0];
		s[0] = '\'';
		s[2] = '\'';
		s[3] = 0;
		node->SetData(s, 4);
		break;
	case OP_ID:
		TInputDialog(window, h, "Enter an identifier name",
						 s, sizeof s).Execute();
		node->SetData(s, strlen(s)+1);
		break;
	}
}

Sort      GetOpSort(Operator op)
{
	if(op<OP_FIRST)
		return (Sort)op;
	switch(op) {
	case OP_PROG: 				return SORT_PROG;

	case OP_DECS: 				return SORT_DECS;

	case OP_PROCDEC:
	case OP_FUNCDEC:
	case OP_NDECS: 			return SORT_DEC;

	case OP_NDEC:				return SORT_NDEC;

	case OP_ID:					return SORT_ID;

	case OP_INTTYPE:
	case OP_BOOLTYPE:
	case OP_CHARTYPE:
	case OP_REALTYPE:			return SORT_TYPE;

	case OP_STATS:				return SORT_STATS;
	case OP_SKIP:
	case OP_CONCASSIGN:
	case OP_PROCCALL:
	case OP_PROCCALLARGS:
	case OP_IFTHEN:
	case OP_IFTHENELSE:
	case OP_WHILE:
	case OP_READ:
	case OP_WRITE:
	case OP_FOR:
	case OP_BLOCK:
	case OP_NUMCASE:
	case OP_LABCASE:			return SORT_STAT;

	case OP_VARS:				return SORT_VARS;

	case OP_EXPRS:				return SORT_EXPRS;

	case OP_VARVALUE:
	case OP_FUNCCALL:
	case OP_FUNCCALLARGS:
	case OP_NUMVALUE:
	case OP_CHARVALUE:
	case OP_IMPLICATION:
	case OP_EQUIVALENCE:
	case OP_OR:
	case OP_AND:
	case OP_LESS:
	case OP_LESSEQ:
	case OP_GREATER:
	case OP_GREATEREQ:
	case OP_EQUALITY:
	case OP_DIFFERENCE:
	case OP_ADDITION:
	case OP_SUBTRACTION:
	case OP_MULTIPLICATION:
	case OP_DIVISION:
	case OP_MODULO:
	case OP_UNARYMINUS:
	case OP_NOT:
	case OP_EXPRBLOCK:			return SORT_EXPR;

	case OP_PBODYSTATS:
	case OP_PBODYPARSSTATS:		return SORT_PBODY;

	case OP_PPARS:					return SORT_PPARS;

	case OP_PPARVALUE:
	case OP_PPARREF:		 		return SORT_PPAR;

	case OP_PARGS:					return SORT_PARGS;

	case OP_PARGVALUE:
	case OP_PARGREF:		 		return SORT_PARG;

	case OP_FBODY:					return SORT_FBODY;

	case OP_FPARS:					return SORT_FPARS;

	case OP_FPARVALUE:			return SORT_FPAR;

	case OP_FARGS:					return SORT_FARGS;

	case OP_ALTS1:					return SORT_ALTS1;

	case OP_ALTS2:					return SORT_ALTS2;

	case OP_ALT2:					return SORT_ALT2;

	case OP_LABELS:				return SORT_LABELS;

	case OP_CHARLABEL:
	case OP_NUMLABEL:				return SORT_LABEL;
	}
	return SORT_NONE;
}



char     *OpName[] = {
	"PROG", "DECS", "NDECS", "NDEC", "PROCDEC", "FUNCDEC",
	"ID", "INT", "BOOL", "CHAR",	"REAL",	"STATS", "SKIP",
	"CONCASSIGN", "PROCCALL", "PROCCALLARGS", "IFTHEN", "IFTHENELSE",
	"WHILE", "READ", "WRITE", "FOR", "BLOCK", "VARS", "EXPRS",
	"VARVALUE", "FUNCCALL", "FUNCCALLARGS", "NUMVALUE", "CHARVALUE",
	"IMPLICATION", "EQUIVALENCE", "OR",  "AND",  "LESS",  "LESSEQ",
	"GREATER",  "GREATEREQ", "EQUALITY", "DIFFERENCE", "ADDITION",
	"SUBTRACTION", "MULTIPLICATION", "DIVISION", "MODULO",
	"UNARYMINUS", "NOT", "EXPRBLOCK", "PBODYSTATS", "PBODYPARSSTATS",
	"PPARS", "PPARVALUE", "PPARREF", "PARGS", "PARGVALUE", "PARGREF",
	"FBODY", "FPARS", "FPARVALUE", "FARGS",
	"NUMCASE", "LABCASE", "ALTS1", "ALTS2", "ALT2",
	"LABELS", "NUMLABEL", "CHARLABEL"
};

char     *SortName[] = {
	"",
	"PROG", "DEC", "NDEC", "DECS",
	"ID",	 "TYPE", "STAT", "STATS", "VARS", "EXPRS", "EXPR",
	"ALTS1","ALTS2", "ALT2", "LABELS", "LABEL",
	"PBODY","PPARS",
	"PPAR", "PARGS", "PARG", "FBODY", "FPARS",
	"FPAR", "FARGS"
};

char   *GetSortName(Sort sort)
{
	return SortName[(int)sort];
}

char   *GetOpName(Operator op)
{
	if(!op)
		return 0;
	if(op<OP_FIRST) {
                static char     nbuf[MAXOPNAMELEN+1];
		wsprintf(nbuf, "%sHOLE", GetSortName((Sort)op));
		return nbuf;
	} else {
		return OpName[(int)(op-OP_FIRST)];
	}
}

