//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TStructDocument
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop
#include "strucdoc.h"



UINT		TStructDocument::ClipboardFormat = 0;

TStructDocument::TStructDocument(TDocument* parent) : TFileDocument(parent)
{
	SetDirty(FALSE);
	ActiveNode = Editor.GetRoot();
}


void     TStructDocument::Refine(TNodeAddress& address, Operator op, TWindow *window)
{
	Editor.Refine(address, op);
	AskData(window, Editor.NodeAt(address));
	ActiveNode = Editor.NodeAt(address);

	SetDirty(TRUE);
	NotifyViews(vnMirrorSubtree, (long)&address);
}


void     TStructDocument::Delete(TNodeAddress& address)
{
	Editor.Delete(address);
	ActiveNode = Editor.NodeAt(address);

	SetDirty(TRUE);
	NotifyViews(vnMirrorSubtree, (long)&address);
}

void     TStructDocument::Cut(TNodeAddress& address, HWND hwnd)
{
	Copy(address, hwnd);
	Delete(address);
}
void     TStructDocument::Copy(TNodeAddress& address, HWND hwnd)
{
	TTerm      *node    = Editor.NodeAt(address);
	int         memlen  = node->GetTermLength()+1;
	HGLOBAL     handle, handle2 = 0;
	char       *p  = 0;
	Sort  	   *p2 = 0;
	TClipboard  cb(hwnd);

	handle  = ::GlobalAlloc(GHND, memlen);
	if(!handle)
		goto quit;
	handle2 = ::GlobalAlloc(GHND, sizeof(Sort));
	if(!handle2)
		goto quit;
	p = (char *)::GlobalLock(handle);
	if(!p)
		goto quit;
	p2 = (Sort *)::GlobalLock(handle2);
	if(!p2)
		goto quit;

	if(!ClipboardFormat) {
		ClipboardFormat = cb.RegisterClipboardFormat("WinStruc Term Sort");
	}
	if(cb.EmptyClipboard()) {
		{
			ostrstream  os(p, memlen);
			Editor.WriteSubtreeToStream(os, address);
			os << '\0';  // write terminating 0
		}
		*p2 = node->GetSort();

		::GlobalUnlock(handle);
		::GlobalUnlock(handle2);
		cb.SetClipboardData(CF_TEXT,         handle);
		cb.SetClipboardData(ClipboardFormat, handle2);
		handle = 0; handle2 = 0; p = 0; p2 = 0;
	}
quit:
	if(p)
		::GlobalUnlock(handle);
	if(p2)
		::GlobalUnlock(handle2);
	if(handle)
		GlobalFree(handle);
	if(handle2)
		GlobalFree(handle2);
}
void     TStructDocument::Paste(TNodeAddress& address, HWND hwnd)
{
	TTerm      	*node    = Editor.NodeAt(address);
	HGLOBAL      handle, handle2;
	char       	*p = 0;
	Sort  		*p2 = 0;
	TClipboard   cb(hwnd);

	if(!ClipboardFormat)
		goto quit;
	handle2 = cb.GetClipboardData(ClipboardFormat);
	if(!handle2)
		goto quit;
	handle  = cb.GetClipboardData(CF_TEXT);
	if(!handle)
		goto quit;
	p2 = (Sort *)::GlobalLock(handle2);
	if(!p2 || *p2 != node->GetSort())
		goto quit;
	p  = (char *)::GlobalLock(handle);
	if(!p)
		goto quit;
	{
		istrstream  is(p);
		Editor.ReadSubtreeFromStream(is, address);
	}
	SetDirty(TRUE);
	ActiveNode = Editor.NodeAt(address);
	NotifyViews(vnMirrorSubtree, (long)&address);
quit:
	if(p)
		::GlobalUnlock(handle);
	if(p2)
		::GlobalUnlock(handle2);
}

BOOL     TStructDocument::ClipboardOK(const Sort sort, HWND hwnd) const
{
	BOOL		IsOK = FALSE;
	if(ClipboardFormat) {
		TClipboard  cb(hwnd);
		HGLOBAL     handle2;

		handle2 = cb.GetClipboardData(ClipboardFormat);
		if(handle2) {
			Sort *p2 = (Sort *)::GlobalLock(handle2);
			if(p2) {
				if(*p2 == sort)
					IsOK = TRUE;
				::GlobalUnlock(handle2);
			}
		}
	}
	return IsOK;
}


void     TStructDocument::ListInsert(TNodeAddress& address, int childnumber)
{
	Editor.ListInsert(address, childnumber);
	SetDirty(TRUE);
	NotifyViews(vnMirrorSubtree, (long)&address);
}


void     TStructDocument::ListDelete(TNodeAddress& address, int childnumber)
{
	Editor.ListDelete(address, childnumber);
	SetDirty(TRUE);
	ActiveNode = Editor.NodeAt(address);
	NotifyViews(vnMirrorSubtree, (long)&address);
}


void     TStructDocument::GlobalFocus(TNodeAddress& address)
{
	SetActiveAddress(address);
	NotifyViews(vnGlobalFocus, (long)&address);
}



BOOL TStructDocument::Commit(BOOL force)
{
	if (!IsDirty() && !force)
		return TRUE;

	TOutStream* os = OutStream(ofWrite);
	if (!os)
		return FALSE;

	Editor.WriteToStream(*os);

	delete os;

	SetDirty(FALSE);
	return TRUE;
}

BOOL TStructDocument::Revert(BOOL clear)
{
	if (!TFileDocument::Revert(clear))
		return FALSE;
	if (!clear)
		Open(0, 0);
	static TNodeAddress rootaddr;
	ActiveNode = Editor.GetRoot();
	NotifyViews(vnMirrorSubtree, (long)&rootaddr);
	return TRUE;
}

BOOL   TStructDocument::Open(int /*mode*/, LPCSTR path)
{
	if (path)
		SetDocPath(path);
	if (GetDocPath()) {
		TInStream* is = InStream(ofRead);
		if (!is)
			return FALSE;

		Editor.ReadFromStream(*is);

		delete is;
	}
	ActiveNode = Editor.GetRoot();
	SetDirty(FALSE);
	return TRUE;
}

void				 TStructDocument::GetActiveAddress(TNodeAddress& address) const
{
	Editor.GetNodeAddress(ActiveNode, address);
}
void				 TStructDocument::SetActiveAddress(TNodeAddress& address)
{
	TTerm		*newactive = Editor.NodeAt(address);

	if(ActiveNode != newactive) {
		ActiveNode = newactive;
		NotifyViews(vnNewActive, (long)&address);
	}
}

