//----------------------------------------------------------------------------
//    Copyright 1995 Hugo Lyppens
//   	Implements class TViewNode
//----------------------------------------------------------------------------
#include "owlhdr.h"
#pragma hdrstop
#include "tree.h"
#include "mydc.h"
#include "strucdoc.h"
#include "viewnode.h"

TViewNode::TViewNode(int n, int fixed):
	TTerm(n, fixed),Folded(FALSE),
        NodePositions(0,0,2)
{}

void        TViewNode::SetNodeSize(TSize& sz)
{  Size = sz;
}
void        TViewNode::AddNodePosition(TPoint& pt)
{	NodePositions.Add(pt);
}

int 		   TViewNode::GetNumNodePositions() const
{	return NodePositions.GetItemsInContainer();
}
TPoint&		TViewNode::GetNodePosition(int n) const
{
	return NodePositions[n];
}


TViewNode        *TViewNode::WhichNode(TPoint& point)
{
	TViewNode     						*foundnode = 0;
	TArrayAsVectorIterator<TPoint> iter(NodePositions);

	while(iter && !foundnode) {
		if(TRect(iter++, NodeSize()).Contains(point)) {
			if(!GetFolded()) {
				int numchildren = GetNumChildren();
				for(int i = 0; !foundnode && i<numchildren; i++)
					foundnode = GetChild(i)->WhichNode(point);
			}
			if(!foundnode)
				foundnode = this;
		}
	}
	return foundnode;
}

BOOL        TViewNode::IsVisible()  const
{
   const TViewNode  const *p = this;

   while(p = p->GetParent()) {
      if(p->GetFolded())
			return FALSE;
	}
	return TRUE;
}

void						TViewNode::FlushChildrenNodePositions()
{
	int			 nc = GetNumChildren();

	for(int i = 0; i<nc; i++) {
		GetChild(i)->NodePositions.Flush();

	}
}


